/************************************************************************************************
 * Copyright  (c) 2011, 2012 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

import SAFETypes::*;

interface AGUIfc;
   	method AGUOut do_ag_check(OPCODE op, AtomicGroup r1_ag, AtomicGroup r2_ag, AtomicGroup r3_ag);
   	method Bool do_mr_ag_check(OPCODE op, AtomicGroup r1_ag, AtomicGroup r2_ag, AtomicGroup r3_ag);
endinterface: AGUIfc

(* synthesize *)
module mkAGU (AGUIfc);

   	method AGUOut do_ag_check (OPCODE op, AtomicGroup r1_ag, AtomicGroup r2_ag, AtomicGroup r3_ag);
      	AGUOut out;
	  	Bool is_error;
	  	AtomicGroup res;

		case(op)
	 		ADD, SUB, OR, AND, SHL, SHR, MUL:
			case(r1_ag)
		     	        INTEGER: is_error = r1_ag != r2_ag;
                                ERROR: is_error = False;
		     	        default: if (r2_ag == ERROR)
                                                 is_error = False;
                                         else
                                                 is_error = True;
		  	endcase
	 		TEST: is_error = r1_ag != r2_ag;
	 		BNEG, BPOS, BEQ, BNE, GTIME: is_error = (r1_ag!=INTEGER);
			CPMR, MVMR:
			case (r1_ag)
				FR_POINTER, CONST_POINTER, LIN_FR_POINTER, 
                                INSTR_POINTER, THREAD_POINTER: is_error = False;
				default: is_error = True;
			endcase
	 		CPRM,MVRM:
			case (r2_ag)
				FR_POINTER, LIN_FR_POINTER, INSTR_POINTER, THREAD_POINTER: is_error = False;
				default: is_error = True;
	 		endcase
                        CPPM, CPMP:
			case (r2_ag)
                                THREAD_POINTER: is_error = False;
				default: is_error = True;
	 		endcase
	 		FRAMPTR: is_error = !((r1_ag == INTEGER) && (r2_ag == FR_POINTER));
	 		OFFP: is_error = !((r1_ag == INTEGER) && ((r2_ag == FR_POINTER) ||
                                                                   (r2_ag == INSTR_POINTER)));
	 		OFFLP: is_error = !((r1_ag == INTEGER) && ((r3_ag == FR_POINTER) ||
                                                                   (r3_ag == INSTR_POINTER) ||
                                                                   (r3_ag == LIN_FR_POINTER) ||
                                                                   (r3_ag == THREAD_POINTER) ||
                                                                   (r3_ag == STREAM_READ_POINTER) ||
                                                                   (r3_ag == STREAM_WRITE_POINTER)));
			OFFTP: is_error = !((r1_ag == INTEGER) && (r3_ag == THREAD_POINTER));
	 		TOTAG: is_error = (r1_ag != TAG);
	 		MAKEP: is_error = (r1_ag != FR_POINTER);
	 		INP: is_error = (r1_ag != PRINCIPAL);
	 		INA: is_error = (r1_ag != AUTHORITY);
	 		SETA: is_error = (r1_ag != PRINCIPAL);
	 		GCALL, GJMP: is_error = (r1_ag != GATE_POINTER);
			GACALL, GAJMP: is_error = (r1_ag != GATE_POINTER) || (r2_ag != AUTHORITY);
	 		CALL: is_error = (r1_ag != INSTR_POINTER);
	 		RUNT, RESUMET: is_error = !(r1_ag == THREAD_POINTER);
	 		STWY, STWB: is_error = (r2_ag != STREAM_WRITE_POINTER);
	 		STRY, STRB: is_error = (r1_ag != STREAM_READ_POINTER);
	 		FJMP: is_error = (r1_ag!=INSTR_POINTER);
	 		GATE: is_error = (r1_ag!=INSTR_POINTER) || (r2_ag!=FR_POINTER) || (r3_ag!=GATE_POINTER);
	 		TMUU, TMURC, NEWG: is_error = (r1_ag!=INTEGER);
	 		TMUL: is_error = !(((r1_ag == INTEGER) && (r2_ag == INTEGER)) && ((r3_ag == FR_POINTER)||(r3_ag == AUTHORITY)||(r3_ag == INTEGER)));
			CPMT, CPIO, CPAR, CPHR: is_error = (r1_ag!=THREAD_POINTER);
			REGRP: is_error = !(r2_ag == r3_ag);
			//MVRR,CPRR,LCFP,NEWT,RETAG,JMP,NOP,HALT,YIELD,GRTN,SETTAG,INTAG,REGRP,INGRP,REMT,RDTIME,TRANSBEGIN: is_error = False;  //AG doesnt matter for these instructions...
	 		default: is_error = False;
		endcase

		case(op)
			ADD, SUB, OR, AND, SHL, SHR, MUL, TOTAG, INGRP, INTAG, GTIME, RDTIME, REMT, STRB, STRY, TEST, TMURC, CPIO, CPHR, TESTGRP:
                        if (r1_ag  ==  ERROR || r2_ag  ==  ERROR)
                                res = ERROR;
                        else
                                res = INTEGER;
	 		BNEG, BPOS: res = ERROR;
	 		MVRR, CPRR: res = r1_ag;
                        LCFP: res = CONST_POINTER;
	 		CPMT, RFLCT: res = FR_POINTER;
	 		FRAMPTR, OFFP, RETAG, SETTAG, NEWT: res = r2_ag;
			OFFLP, OFFTP: res = r3_ag;
	 		MAKEP: res = PRINCIPAL;
	 		INP, INA: res = FR_POINTER;
	 		REGRP, SETA, STWY, STWB: res = r1_ag;
	 		GATE: res = GATE_POINTER;
			TAGOF: res = TAG;
			CPAR: res = AUTHORITY;
			BRTN: res = r1_ag;
	 		default: res = UNINITIALIZED;
      	        endcase

	  	out = tuple2(is_error,res);

		return out;

	endmethod: do_ag_check

	method Bool do_mr_ag_check(OPCODE op, AtomicGroup ag1, AtomicGroup ag2, AtomicGroup ag3);
		Bool is_error;
		case(op)
			GCALL, GACALL: is_error = !((ag1 == INSTR_POINTER) && (ag2 == AUTHORITY));
		        default: is_error = False;
		endcase

		return is_error;
	endmethod

endmodule: mkAGU
