/************************************************************************************************
* Copyright  (c) 2011, 2012 University of Pennsylvania
*
* Permission to use, copy, modify, and distribute this software and
* its documentation for any purpose, without fee, and without a
* written agreement is hereby granted, provided that the above copyright
* notice and this paragraph and the following two paragraphs appear in
* all copies.
*
* IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
* DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
* HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
* AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
* AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
* PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
***********************************************************************************************/

import SAFETypes::*;

interface ALUIfc;
        method ALUOut do_alu (OPCODE op, Payload src1, Payload src2);
endinterface: ALUIfc

(*synthesize*)
module mkALU (ALUIfc);

        method ALUOut do_alu (OPCODE op, Payload src1, Payload src2);
        //change the operands to an interger type for calculation
                Int#(PAYLOAD_WIDTH) operand1 = unpack(src1);
                Int#(PAYLOAD_WIDTH) operand2 = unpack(src2);
	        Int#(PAYLOAD_WIDTH) default_value = unpack(0);
                case (op)
	                ADD: return tuple2(False, pack(operand1 + operand2));
	                SUB: return tuple2(False, pack(operand1 - operand2));
	                AND: return tuple2(False, pack(operand1 & operand2));
	                OR: return tuple2(False, pack(operand1 | operand2));
	                SHL: return tuple2(False, pack(operand1 << operand2));
	                SHR: return tuple2(False, pack(operand1 >> operand2));
                        //MUL: return tuple2(False, pack(operand1 * operand2));
	                TEST: begin
		                      Bit#(PTR_WIDTH) op1_short = truncate(pack(operand1));
		                      Bit#(PTR_WIDTH) op2_short = truncate(pack(operand2));

	 	                      if(op1_short == op2_short)
			                      begin
				                      return tuple2(False, unpack(1));
			                      end
		                      else return tuple2(False, unpack(0));
		              end
	                default: return tuple2(False, pack(default_value)); //if not a valid instruction, return don't care
                endcase
        endmethod: do_alu

endmodule: mkALU
