/************************************************************************************************
 * Copyright  (c) 2011, 2012 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

import SAFETypes::*;

interface FatPUIfc;
   	method Tuple3#(FatPointer, Dunder, Dover) update_pointer(FatPointer ptr, Dunder dunder, Dover dover, A offset, Bool is_stream);
	method Tuple2#(Dunder, Dover) decode(FatPointer ptr);
endinterface: FatPUIfc

(*synthesize*)
module mkFatPU (FatPUIfc);

	method Tuple3#(FatPointer, Dunder, Dover) update_pointer(FatPointer ptr, Dunder dunder, Dover dover, A offset, Bool is_stream);
                Dunder tmp_dunder = dunder + unpack(offset);
                Dover tmp_dover = dover - unpack(offset);
                Dunder dest = unpack(ptr.a) + unpack(offset);
                A a_1 = pack(unpack(ptr.a) - dunder);
                //do wrap-around buffer
                if (is_stream && (dover == 0))
                        return tuple3(FatPointer{meta: MetaData{ag: ptr.meta.ag,tag: ptr.meta.tag},
                                                 a: a_1, b: ptr.b, i: ptr.i, m: ptr.m},
                                      0, dunder);
                else if (tmp_dunder < 0 || tmp_dover < 0)
                        return tuple3(FatPointer{meta: MetaData{ag: ERROR, tag: ptr.meta.tag},
                                                 a: 0, b: 0, i: 0, m: 0}, 
                                      tmp_dunder, tmp_dover);
                else
                        return tuple3(FatPointer{meta: MetaData{ag: ptr.meta.ag,tag: ptr.meta.tag},
                                                 a: pack(dest), b: ptr.b, i: ptr.i, m: ptr.m},
                                      tmp_dunder, tmp_dover);
	endmethod

        method Tuple2#(Dunder, Dover) decode (FatPointer ptr);
                Bit#(6) new_b = truncate(ptr.b);
                A dunder_1;
                A dover_1;
                A dunder_2;
                A dover_2;
                
                A c = (1 << 6) << new_b;
                A m = extend(c - 1);
                
                A a = ptr.a >> new_b;
                A i = extend(ptr.i) << new_b;
                A u = extend(ptr.m) << new_b;
                
                dunder_1 = (c  | (m & ptr.a)) - i;
                dunder_2 = (m & ptr.a) - i;
                
                dover_1 = (c | u) - (m & ptr.a) - 1;
                dover_2 = u - (m & ptr.a) - 1;
                        
                return tuple2(ptr.i > a[5:0] ? unpack(dunder_1) : unpack(dunder_2), 
                              ptr.m < a[5:0] ? unpack(dover_1) : unpack(dover_2));
        endmethod

endmodule