/************************************************************************************************
 * Copyright  (c) 2011, 2012 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

import SAFETypes::*;

interface GSIfc;
   	method FatPointer get_gsp;
   	method Action set_gsp (FatPointer new_gsp, Dunder new_dunder, Dover new_dover);
	method Action update_gsp(OPCODE op);
        method Dunder get_gsp_dunder;
        method Dover get_gsp_dover;
endinterface: GSIfc

(*synthesize*)
module mkGateStack(GSIfc);

   	Reg#(FatPointer) gsp <- mkRegU;
        Reg#(Dunder) dunder <- mkRegU;
        Reg#(Dover) dover <- mkRegU;
	Reg#(A) base <- mkRegU;
	Reg#(A) twotobm1 <- mkRegU;
	Reg#(A) temp1 <- mkRegU;
	Reg#(A) temp2 <- mkRegU;

   	method FatPointer get_gsp;
		return gsp;
   	endmethod

	method Action update_gsp(OPCODE op);
		FatPointer temp = gsp;
                Dunder new_dunder;
                Dover new_dover;

		case(op)
			GCALL, GACALL, CALL, BCALL, BGCALL, BGACALL:
                        begin
				temp.a = gsp.a + 4;
                                new_dunder = dunder + 4;
                                new_dover = dover - 4;
			end
			GRTN, BRTN:
                        begin
				temp.a = gsp.a - 4;
                                new_dunder = dunder - 4;
                                new_dover = dover + 4;
			end
			default:
                        begin
                                temp.a = gsp.a;
                                new_dunder = dunder;
                                new_dover = dover;
                        end
		endcase

                gsp <= temp;
                dunder <= new_dunder;
                dover <= new_dover;
                //$display("GSP new dunder: %x, dover: %x", new_dunder, new_dover);
	endmethod

        method Dunder get_gsp_dunder;
                return dunder;
	endmethod

	method Dover get_gsp_dover;
                return dover;
	endmethod


   	method Action set_gsp(FatPointer new_gsp, Dunder new_dunder, Dover new_dover);
                dunder <= new_dunder;
                dover <= new_dover;
		gsp <= new_gsp;
                //$display("setting GSP. gsp: %x, dunder: %x, dover: %x", new_gsp, new_dunder, new_dover);
   	endmethod

endmodule: mkGateStack



