/*************************************************
hash function module for TMU
Author: Udit Dhawan
IC Group, University of Pennsylvania

Last update: 11.08.2011
Last Change: made hash functions "truly" parameterizable
**************************************************/

import SAFETypes::*;

interface HashIfc;
	method Bit#(9) compute_hash(TMU_M tmu_m); //tmu_m =  {tptag, citag, opgrp, pctag, auth, op1tag, op2tag, op3tag, mrtag}
endinterface: HashIfc

(* synthesize *)
module mkHash #(parameter int g_addrlen, int s1, int s2, int s3, int s4, int s5, int s6, int s7, int s8, int s9) (HashIfc);

	function TAG_A rotate(TAG_A intag, int r);
		TAG_A out_tag;

		out_tag = (intag<<r) ^ (intag>>(28-r));
		return out_tag;
	endfunction

	function Bit#(9) fold_collapse(TAG_A hash_val);
		Bit#(9) sample1;
		Bit#(9) sample2;
		Bit#(9) sample3;
		Bit#(9) sample4;

		sample1 = truncate(pack(hash_val));
		sample2 = truncate(pack(hash_val)>>9);
		sample3 = truncate(pack(hash_val)>>18);
		sample4 = truncate(pack(hash_val)>>27);

		Bit#(9) res = truncate(sample1 ^ sample2 ^ sample3 ^ sample4);

		return res;
	endfunction

	method Bit#(9) compute_hash(TMU_M tmu_m);

		let tmp = fold_collapse(rotate(tmu_m.tptag,s1)) ^ fold_collapse(rotate(tmu_m.citag,s2)) ^ fold_collapse(rotate(unpack(extend(pack(tmu_m.opgrp))),s3)) ^ fold_collapse(rotate(tmu_m.pctag,s4)) ^ fold_collapse(rotate(tmu_m.auth,s5)) ^ fold_collapse(rotate(tmu_m.op1tag,s6)) ^ fold_collapse(rotate(tmu_m.op2tag,s7)) ^ fold_collapse(rotate(tmu_m.op3tag,s8)) ^ fold_collapse(rotate(tmu_m.mrtag,s9));

		Bit#(9) hash_val = truncate(tmp);

		return hash_val;

	endmethod

endmodule: mkHash
