/**********************************************
Interposed Unit to support arbitrary number of streams to and from the host
Authors: Albert Kwon, Alen Kubati
IC Group, University of Pennsylvania

Last Updated: 08.xx.2011
Last Change:
***********************************************/

import SAFETypes::*;
import BRAM::*;

interface InterposedIfc;
        method Bool do_stream(Bit#(2) offset, Bool write, CacheLine value);
        method BRAMRequest#(A, CacheLine) do_load_dmem(A addr, CacheLine atom);
        method BRAMRequest#(A, CacheLine) do_write(A addr, Payload pay);
        method BRAMRequest#(A, CacheLine) read_request(A addr);
        method BRAMRequest#(A, CacheLine) do_read(A addr);
endinterface: InterposedIfc

(* synthesize *)
module mkIU (InterposedIfc);

        method Bool do_stream(Bit#(2) offset, Bool write, CacheLine value);
                Atom mem_val;
                if (offset == 0) begin
	                mem_val = unpack(pack(value.a0));
                end
                else if (offset == 1) begin
	                mem_val = unpack(pack(value.a1));
                end
                else if (offset == 2) begin
	                mem_val = unpack(pack(value.a2));
                end
                else begin
	                mem_val = unpack(pack(value.a3));
                end
                return (write && (mem_val.meta.ag != EMPTY)) || (!write && (mem_val.meta.ag == EMPTY));
        endmethod

	method BRAMRequest#(A, CacheLine) do_load_dmem(A addr, CacheLine atom);
      	        return BRAMRequest {write: True, responseOnWrite: False, address: addr, datain: atom};
   	endmethod

   	method BRAMRequest#(A, CacheLine) do_write(A addr, Payload pay);
      	        return BRAMRequest {write: True, responseOnWrite: False, address: addr, datain: unpack(0)};
   	endmethod

   	method BRAMRequest#(A, CacheLine) read_request(A addr);
      	        return BRAMRequest {write: False, responseOnWrite: False, address: addr, datain: ?};
   	endmethod

   	method BRAMRequest#(A, CacheLine) do_read(A addr);
      	        return BRAMRequest {write: True, responseOnWrite: False, address: addr, datain: unpack(0)};
        endmethod

endmodule: mkIU
