/************************************************************************************************
 * Copyright  (c) 2011, 2012 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

import SAFETypes::*;
import BRAM::*;
import GetPut::*;
import ClientServer::*;

interface LSUIfc;
   method BRAMRequest#(A, CacheLine) read_request(OPCODE op, A addr);
   method BRAMRequest#(A, CacheLine) do_store(OPCODE op, A addr, CacheLine src);
endinterface: LSUIfc

(*synthesize*)
module mkLSU(LSUIfc);

   method BRAMRequest#(A, CacheLine) read_request(OPCODE op, A addr);
      case(op)
	 STWB, STWY, CPMR, CPRM, CPMT, CPIO, CPAR, CPHR, STRB, STRY, MVMR, MVRM, GCALL, GACALL, GAJMP, GJMP, GRTN, CPMP, CPPM: return BRAMRequest {write: False, responseOnWrite: False, address: addr, datain: ?};
	 default: return BRAMRequest {write: False, responseOnWrite: False, address: ?, datain: ?};
      endcase
   endmethod: read_request

   method BRAMRequest#(A, CacheLine) do_store(OPCODE op, A addr, CacheLine src);
      case(op)
	 CPRM, STRB, STRY, STWB, STWY, MVMR, MVRM, GCALL, GACALL, CALL, GATE, CPPM: return BRAMRequest {write: True, responseOnWrite: False, address: addr, datain: src};
	 default: return BRAMRequest {write: False, responseOnWrite: False, address: ?, datain: ?};
      endcase
   endmethod: do_store

endmodule: mkLSU
