/************************************************************************************************
 * Copyright  (c) 2011, 2012 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

import SAFETypes::*;

import BRAMCore::*;
import BRAM::*;
import ClientServer::*;
import GetPut::*;

interface MemIfc;
   interface Server#(BRAMRequest#(A, CacheLine), CacheLine) rw;
   interface Server#(BRAMRequest#(A, CacheLine), CacheLine) proc_rw;
   interface Put#(BRAMRequest#(A, CacheLine)) store;
endinterface

(*synthesize*)
module mkMem(MemIfc);

   	BRAM_DUAL_PORT#(A, Atom) dmem0 <- mkBRAMCore2Load(4096, False, "mem0.img", False);
	BRAM_DUAL_PORT#(A, Atom) dmem1 <- mkBRAMCore2Load(4096, False, "mem1.img", False);
	BRAM_DUAL_PORT#(A, Atom) dmem2 <- mkBRAMCore2Load(4096, False, "mem2.img", False);
	BRAM_DUAL_PORT#(A, Atom) dmem3 <- mkBRAMCore2Load(4096, False, "mem3.img", False);

        interface Server rw;
                interface Put request;
	                method Action put(req);
                                dmem0.a.put(req.write, req.address, req.datain.a0);
                                dmem1.a.put(req.write, req.address, req.datain.a1);
                                dmem2.a.put(req.write, req.address, req.datain.a2);
                                dmem3.a.put(req.write, req.address, req.datain.a3);
	                endmethod
                endinterface
                interface Get response;
	                method ActionValue#(CacheLine) get;
                                let tmp0 = dmem0.a.read;
                                let tmp1 = dmem1.a.read;
                                let tmp2 = dmem2.a.read;
                                let tmp3 = dmem3.a.read;
	                        return CacheLine{a0: tmp0, a1: tmp1, a2: tmp2, a3: tmp3};
	                endmethod
                endinterface
        endinterface

        interface Server proc_rw;
                interface Put request;
                        method Action put(req);
                                dmem0.b.put(req.write, req.address, req.datain.a0);
                                dmem1.b.put(req.write, req.address, req.datain.a1);
                                dmem2.b.put(req.write, req.address, req.datain.a2);
                                dmem3.b.put(req.write, req.address, req.datain.a3);
	                endmethod
                endinterface
                interface Get response;
	                method ActionValue#(CacheLine) get;
                                let tmp0 = dmem0.b.read;
                                let tmp1 = dmem1.b.read;
                                let tmp2 = dmem2.b.read;
                                let tmp3 = dmem3.b.read;
	                        return CacheLine{a0: tmp0, a1: tmp1, a2: tmp2, a3: tmp3};
	                endmethod
                endinterface
        endinterface

        interface Put store;
                method Action put(req);
                        dmem0.b.put(req.write, req.address, req.datain.a0);
                        dmem1.b.put(req.write, req.address, req.datain.a1);
                        dmem2.b.put(req.write, req.address, req.datain.a2);
                        dmem3.b.put(req.write, req.address, req.datain.a3);
                endmethod
        endinterface

endmodule
