/************************************************************************************************
 * Copyright  (c) 2011, 2012 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

import SAFETypes::*;

interface PCFatPUIfc;
   	method Tuple3#(FatPUOut, Dunder, Dover) update_pointer(FatPointer ptr, Dunder dunder, Dover dover, A offset, Bool from_base);
	method Tuple3#(Dunder, Dover, A) decode(FatPointer ptr);
endinterface

(*synthesize*)
module mkPCFatPU (PCFatPUIfc);

	method Tuple3#(FatPUOut, Dunder, Dover) update_pointer(FatPointer ptr, Dunder dunder, Dover dover, A offset, Bool from_base);
                Int#(PTR_WIDTH) diff = from_base ? unpack(offset) - dunder : unpack(offset);
                let tmp_dunder = dunder + diff;
                let tmp_dover = dover - diff;
                Int#(PTR_WIDTH) dest = unpack(ptr.a) + diff;
                //do wrap-around buffer                
                if (tmp_dunder < 0 || tmp_dover < 0)
                        return tuple3(tuple2(False, FatPointer{meta: MetaData{ag: ERROR, tag: ptr.meta.tag},
                                                              a: 0, b: 0, i: 0, m: 0}), tmp_dunder, tmp_dover);
                else
                        return tuple3(tuple2(True, FatPointer{meta: MetaData{ag: ptr.meta.ag,tag: ptr.meta.tag},
                                                              a: pack(dest), b: ptr.b, i: ptr.i, m: ptr.m}),
                                      tmp_dunder, tmp_dover);
	endmethod

        method Tuple3#(Dunder, Dover, A) decode (FatPointer ptr);
                Bit#(6) new_b = truncate(ptr.b);

                A a = ptr.a >> new_b;
                A i = extend(ptr.i) << new_b;
                A u = extend(ptr.m) << new_b;
                A c = (1 << new_b) << 6;
                A m = extend(c - 1);

                A dunder;
                A dover;

                if (ptr.i > a[5:0])
                        dunder = (c  | (m & ptr.a)) - i;
                else
                        dunder = (m & ptr.a) - i;

                if (ptr.m <= a[5:0])
                        dover = (c | u) - (m & ptr.a);
                else
                        dover = u - (m & ptr.a);
                Int#(PTR_WIDTH) tmp_base = unpack(0); //unpack(ptr.a) - dunder;
                return tuple3(unpack(dunder), unpack(dover), pack(tmp_base));
        endmethod

endmodule