/************************************************************************************************
 * Copyright  (c) 2011, 2012 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

import SAFETypes::*;
import PCFPU::*;

interface PCUIfc;
        method Action update_pc(A offset, Bool from_base);
        method FatPointer get_pc();
        method Bool pc_ag_check();
        method Dunder get_pc_dunder();
        method Dover get_pc_dover();
        method FatPointer get_pc_base();
        method Action retag_pc(Tag new_tag);
        method Action set_pc();
        method Action compute_dist(FatPointer new_pc);
        method Bool in_bounds();
	method Action load_pc(FatPointer new_pc);
        method Action change_pc(FatPointer new_pc, Dunder new_dunder, Dover new_dover);
endinterface: PCUIfc

(*synthesize*)
module mkPCU(PCUIfc);

        PCFatPUIfc fpu <- mkPCFatPU;

        //instantiate PC
        Reg#(FatPointer) pc <- mkRegU;
        Reg#(Dunder) dunder <- mkRegU;
        Reg#(Dover) dover <- mkRegU;
        Reg#(Bool) in_bound <- mkRegU;
        Reg#(FatPointer) pc_base <- mkRegU;

        Reg#(Dunder) dunder_tmp <- mkRegU;
        Reg#(Dover) dover_tmp <- mkRegU;
        Reg#(FatPointer) pc_base_tmp <- mkRegU;
        Reg#(FatPointer) pc_tmp <- mkRegU;

        method Action update_pc(A offset, Bool from_base);
                let fpu_out = fpu.update_pointer(pc, dunder, dover, offset, from_base);
                pc <= tpl_2(tpl_1(fpu_out));
                in_bound <= tpl_1(tpl_1(fpu_out));
                dunder <= tpl_2(fpu_out);
                dover <= tpl_3(fpu_out);
                //$display("PC in bound: %x, %x, %x, %x, %x", tpl_1(tpl_1(fpu_out)), tpl_2(tpl_1(fpu_out)), tpl_2(fpu_out), tpl_3(fpu_out), offset);
        endmethod: update_pc

        method Bool in_bounds();
                return in_bound;
        endmethod

        method Dunder get_pc_dunder();
                return dunder;
        endmethod

        method Dover get_pc_dover();
                return dover;
        endmethod

        method FatPointer get_pc_base();
                return pc_base;
        endmethod

        method FatPointer get_pc();
                return pc;
        endmethod

        method Bool pc_ag_check();
                return (pc.meta.ag != ERROR) && (pc.meta.ag != INSTR_POINTER);
        endmethod

        method Action retag_pc(Tag new_tag);
                pc.meta.tag <= new_tag;
        endmethod

        //commit the pc
        method Action set_pc();
                pc <= pc_tmp;
                pc_base <= pc_base_tmp;
                dunder <= dunder_tmp;
                dover <= dover_tmp;
                in_bound <= pc_tmp.meta.ag != ERROR;
        endmethod

        //decode the pc
        method Action compute_dist(FatPointer new_pc);
                let ret = fpu.decode(new_pc);
                dunder_tmp <= tpl_1(ret);
                dover_tmp <= tpl_2(ret);
                pc_base_tmp <= FatPointer{meta: new_pc.meta, b: new_pc.b, i: new_pc.i, m: new_pc.m, a: tpl_3(ret)};
                pc_tmp <= new_pc;
                //$display("new pc: %x, dunder: %x, dover: %x, base: %x", new_pc, tpl_1(ret), tpl_2(ret), tpl_3(ret));
        endmethod

        //overwrite the pc
        method Action change_pc(FatPointer new_pc, Dunder new_dunder, Dover new_dover);
                pc <= new_pc;
                dunder <= new_dunder;
                dover <= new_dover;
                in_bound <= new_pc.meta.ag != ERROR;
        endmethod

        //decode and load the pc; only used to load pc from the host
        method Action load_pc(FatPointer new_pc);
                let ret = fpu.decode(new_pc);
                dunder <= tpl_1(ret);
                dover <= tpl_2(ret);
                pc_base <= FatPointer{meta: new_pc.meta, b: new_pc.b, i: new_pc.i, m: new_pc.m, a: tpl_3(ret)};
                pc <= new_pc;
                in_bound <= new_pc.meta.ag != ERROR;
        endmethod

endmodule
