/************************************************************************************************
* Copyright  (c) 2011, 2012 University of Pennsylvania
*
* Permission to use, copy, modify, and distribute this software and
* its documentation for any purpose, without fee, and without a
* written agreement is hereby granted, provided that the above copyright
* notice and this paragraph and the following two paragraphs appear in
* all copies.
*
* IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
* DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
* HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
* AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
* AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
* PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
***********************************************************************************************/


/**********************************************
Default register file for SAFE processor
Authors: Albert Kwon
IC Group, University of Pennsylvania

Last Updated: 03.27.2012
***********************************************/

import SAFETypes::*;

import BRAMCore::*;

interface RegFileIfc;
        method Action read_request(ReadIndices addr);
        method RFOut read();
        method Action write_port1(RF_A addr, RegAtom atom);
        method Action write_port2(RF_A addr, RegAtom atom);
endinterface

(*synthesize*)
module mkRegFile(RegFileIfc);

        BRAM_DUAL_PORT#(RF_A, RegAtom) rf1 <- mkBRAMCore2(48, False);
        BRAM_DUAL_PORT#(RF_A, RegAtom) rf2 <- mkBRAMCore2(48, False);

        method Action read_request(ReadIndices addr);
                rf1.a.put(False, addr[0], ?);
                rf1.b.put(False, addr[1], ?);
                rf2.a.put(False, addr[2], ?);
                rf2.b.put(False, addr[3], ?);
        endmethod

        method RFOut read();
                return RFOut{val0: rf1.a.read, val1: rf1.b.read, val2: rf2.a.read, val3: rf2.b.read};
        endmethod

        method Action write_port1(RF_A addr, RegAtom atom);
                rf1.a.put(True, addr, atom);
                rf2.a.put(True, addr, atom);
                $display("write to port 1. %x, %x", addr, atom);
        endmethod

        method Action write_port2(RF_A addr, RegAtom atom);
                rf1.b.put(True, addr, atom);
                rf2.b.put(True, addr, atom);
                $display("write to port 2. %x, %x", addr, atom);
        endmethod

endmodule
