/************************************************************************************************
* Copyright  (c) 2011, 2012 University of Pennsylvania
*
* Permission to use, copy, modify, and distribute this software and
* its documentation for any purpose, without fee, and without a
* written agreement is hereby granted, provided that the above copyright
* notice and this paragraph and the following two paragraphs appear in
* all copies.
*
* IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
* DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
* HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
* AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
* AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
* PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
***********************************************************************************************/


import SAFETypes::*;
import MPRF::*;
import LVT::*;
import Vector::*;

interface RegFileMultiIfc;
	method Action read_request(ReadIndices read_indices);
	method RFOut read();
	method Action write_port1(RF_A addr, RegAtom val);
        method Action write_port2(RF_A addr, RegAtom val);
endinterface

(*synthesize*)
module mkRegFileMulti(RegFileMultiIfc);

	MPRFIfc mprf <- mkMPRF;
	LVTIfc lvt <- mkLVT;

	Reg#(LVTOut) lvt_out <- mkRegU;

        Wire#(Bool) wen1 <- mkDWire(False);
        Wire#(RF_A) addr1 <- mkDWire(0);
        Wire#(Bool) wen2 <- mkDWire(False);
        Wire#(RF_A) addr2 <- mkDWire(0);
        
        rule update_lvt;
                lvt.update_lvt(addr1, wen1, addr2, wen2);
        endrule
        
	method Action read_request(ReadIndices read_indices);
	        //look-up LVT
		let lvt_out_tmp = lvt.lvt_lookup(read_indices);
		lvt_out <= lvt_out_tmp;

	        //issue request
		mprf.read_request(read_indices[0], read_indices[1], read_indices[2], read_indices[3]);
	endmethod

	method RFOut read();
		RFOut rf_bank0_out;
		RFOut rf_bank1_out;

		RFOut rf_out = ?;
		Tuple2#(RFOut, RFOut) mprf_out;

		mprf_out = mprf.do_read;

		rf_bank0_out = tpl_1(mprf_out);
		rf_bank1_out = tpl_2(mprf_out);
	
		if (lvt_out[0] == 0)
			rf_out.val0 = rf_bank0_out.val0;
		else
			rf_out.val0 = rf_bank1_out.val0;
		if (lvt_out[1] == 0)
			rf_out.val1 = rf_bank0_out.val1;
		else
			rf_out.val1 = rf_bank1_out.val1;
		if (lvt_out[2] == 0)
			rf_out.val2 = rf_bank0_out.val2;
		else
			rf_out.val2 = rf_bank1_out.val2;
		if (lvt_out[3] == 0)
			rf_out.val3 = rf_bank0_out.val3;
		else
			rf_out.val3 = rf_bank1_out.val3;
		return rf_out;
	endmethod

	method Action write_port1(RF_A addr, RegAtom val);
		//send write request
		mprf.do_write_port1(addr, val);
                wen1 <= True;
                addr1 <= addr;
	endmethod
        
        method Action write_port2(RF_A addr, RegAtom val);
	        //send write request
		mprf.do_write_port2(addr, val);
		wen2 <= True;
                addr2 <= addr;
	endmethod
        
endmodule
