/************************************************************************************************
 * Copyright  (c) 2011, 2012 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

import SAFETypes::*;
import PCU::*;
import FatPU::*;
import LSU::*;
import Timer::*;
import GS::*;
import ALU::*;
import BU::*;
import AGU::*;
import SU::*;
import TMU::*;
import WriteBuffer::*;
import RegisterFile::*;
import RegisterFileMulti::*;

import BRAMCore::*;
import BRAM::*;
import GetPut::*;
import ClientServer::*;
import Vector::*;

interface SAFEIfc;
        method Atom get_state(Bit#(32) control);
        method Action enable_proc();
	method Action clear_breakpoint();
        method Action set_pc(FatPointer init_pc);
        method Action set_bp_pc(A bp_pc);
        method Action set_bp_cycle(int bp_cycle);
        method Action load_insn(A address, CacheLine insn);
        method Action load_reg(RF_A address, Atom val);
        method Action load_ucode(A address, Atom val);        
        method Action start();
        method Bool halted();
        method Action get_insn_req(A address);
        method Action get_reg_req(RF_A address);
        method CacheLine get_insn();
        method Atom get_reg();
        interface Client#(BRAMRequest#(A, CacheLine), CacheLine) mem_rw;
        interface Get#(BRAMRequest#(A, CacheLine)) store;
        method Action load_hash_mem(Bit#(3) which, Bit#(9) addr, TMU_MR val);
        method Action unload_tmu(Bit#(8) addr);
	method Action set_dc_tag(TAG_A new_tag);
endinterface

(*synthesize*)
module mkSAFE(SAFEIfc);

        //instantiate all functional units
        PCUIfc pcu <- mkPCU;
        GSIfc gate_stack <- mkGateStack;
        ALUIfc alu <- mkALU;
        LSUIfc lsu <- mkLSU;
        FatPUIfc fpu <- mkFatPU;
        BUIfc bu <- mkBU;
        AGUIfc agu <- mkAGU;
        SUIfc su <- mkSU;
        TimerIfc timer <- mkTimer(100000);	//QT set arbitrarily at the moment...
        TMUIfc tmu <- mkTMU;
        WriteBufferIfc write_buffer <- mkWriteBuffer;
        //RegFileIfc rf <- mkRegFile;
	RegFileMultiIfc rf <- mkRegFileMulti;

        //BRAM for instruction memory
        BRAM_DUAL_PORT#(A, Atom) imem0 <- mkBRAMCore2Load(4096, False, "mem0.img", False);
        BRAM_DUAL_PORT#(A, Atom) imem1 <- mkBRAMCore2Load(4096, False, "mem1.img", False);
        BRAM_DUAL_PORT#(A, Atom) imem2 <- mkBRAMCore2Load(4096, False, "mem2.img", False);
        BRAM_DUAL_PORT#(A, Atom) imem3 <- mkBRAMCore2Load(4096, False, "mem3.img", False);

        BRAM_DUAL_PORT#(A, Atom) ucode <- mkBRAMCore2Load(2048, False, "ucode.img", False);

        //Thread State variables

        //0:pc
        //1:auth
        //2:ts
        //3:gsp
        //4:ci
        //5:op1
        //6:op2
        //7:op3
        //8:mr
        //9:tmur
        //10:fpc
        //11:fa
        //12:h1
        //13:h2
        //bootloader thread has DC
        //all have empty at the end
        Vector#(17, Reg#(Atom)) state <- replicateM(mkRegU);
        FatPointer pc = unpack(pack(state[valueOf(PC)]));
        ThreadStatus ts = unpack(truncate(pack(state[valueOf(TS)])));
        FatPointer auth = unpack(pack(state[valueOf(AUTH)]));
        Atom mr = state[valueOf(MR)];
        FatPointer fA = unpack(pack(state[valueOf(FA)]));
        FatPointer fPC = unpack(pack(state[valueOf(FPC)]));
        Bit#(9) h1 = truncate(pack(state[valueOf(H1)]));
        Bit#(9) h2 = truncate(pack(state[valueOf(H2)]));
        FatPointer tp = unpack(pack(state[valueOf(TP)]));
        //dunder/dover for current tp

	Reg#(Bit#(NumGPR)) live_mask <- mkReg(32'hffffffff);
	Reg#(Bit#(NumGPR)) write_mask <- mkReg(32'hffffffff);
	Reg#(Bool) op1_is_live <- mkReg(False);
	Reg#(Bool) op2_is_live <- mkReg(False);
	Reg#(Bool) op3_is_live <- mkReg(False);
	Reg#(Tag) public_tag <- mkReg(unpack('h58));

        Reg#(Atom) op1 <- mkRegU;
        Reg#(Atom) op2 <- mkRegU;
        Reg#(Atom) op3 <- mkRegU;
        Reg#(Instruction) ci <- mkRegU;

        Reg#(Dunder) dunder_tp <- mkRegU;
        Reg#(Dover) dover_tp <- mkRegU;

        //dunder/over for new pc stored to switch pc from ucode to normal pc
        Reg#(Dunder) dunder_pc <- mkRegU;
        Reg#(Dover) dover_pc <- mkRegU;

        Reg#(OPCODE) op <- mkRegU;                      //current opcode

        Reg#(FatPointer) fTP <- mkRegU;                 //faulting thread pointer
        Reg#(Dunder) dunder_ftp <- mkRegU;
        Reg#(Dover) dover_ftp <- mkRegU;
        
        Reg#(FatPointer) old_tp <- mkRegU;

        //dunder/over for new tp in runt/resumet
        Reg#(Dunder) op1_dunder <- mkRegU;
        Reg#(Dover) op1_dover <- mkRegU;

        //new tp used for cpmp loading; incremented through out the use
        Reg#(FatPointer) tmp_tp <- mkRegU;
        Reg#(Dunder) dunder_tmp_tp <- mkRegU;
        Reg#(Dover) dover_tmp_tp <- mkRegU;

        //the original new tp
        Reg#(FatPointer) new_tp <- mkRegU;
        Reg#(Dunder) dunder_new_tp <- mkRegU;
        Reg#(Dover) dover_new_tp <- mkRegU;

        //micro codo variables
        Reg#(Bool) use_ucode <- mkReg(False);
        Reg#(FatPointer) proc_auth <- mkReg(unpack('hfc));

        //need actual tags and addr; hardware global for ucode
        FatPointer ucode_runt_pc = unpack(0);
        Dunder ucode_runt_dunder = unpack(0);
        Dover ucode_runt_dover = unpack(0);

        FatPointer ucode_resumet_pc = FatPointer{meta: MetaData{ag: INSTR_POINTER, tag: unpack('h7a)}, m: 0, b: 0, i: 0, a: 'h23c};
        Dunder ucode_resumet_dunder = unpack(0);
        Dover ucode_resumet_dover = unpack(130);

        FatPointer ucode_yield_pc = unpack(0);
        Dunder ucode_yield_dunder = unpack(0);
        Dover ucode_yield_dover = unpack(0);

        FatPointer ucode_yield2_pc = unpack(0);
        Dunder ucode_yield2_dunder = unpack(0);
        Dover ucode_yield2_dover = unpack(0);

        FatPointer ucode_flush_pc = unpack(0);
        Dunder ucode_flush_dunder = unpack(0);
        Dover ucode_flush_dover = unpack(0);

        FatPointer ucode_fault_pc = FatPointer{meta: MetaData{ag: INSTR_POINTER, tag: unpack('h7a)}, m: 0, b: 0, i: 0, a: 'h1b8};
        Dunder ucode_fault_dunder = unpack(0);
        Dover ucode_fault_dover = unpack(129);

        FatPointer ucode_boot_pc = FatPointer{meta: MetaData{ag: INSTR_POINTER, tag: unpack('h7a)}, m: 63, b: 0, i: 0, a: 'h174};
        Dunder ucode_boot_dunder = unpack(0);
        Dover ucode_boot_dover = unpack(65);

        //dunder/dover from rf
        Reg#(Dunder) dunder <- mkReg(1);
        Reg#(Dover) dover <- mkReg(1);

        //dunder/dover from memory
        Reg#(Dunder) dunder_mem <- mkReg(1);
        Reg#(Dunder) dover_mem <- mkReg(1);

        Reg#(Bool) pc_from_base <- mkRegU; //branch needs

        //transaction variables
        Reg#(Bit#(16)) trans_id <- mkRegU;	//transaction id
        Reg#(Bool) in_trans <- mkReg(False);
        Reg#(FatPointer) trans_pc <- mkRegU;

        Reg#(Bool) ren <- mkRegU;        //read enable for mem
        Wire#(Bool) ren_wire <- mkBypassWire();

        //TODO: find the right bounds for this on new fatptr
        Reg#(FatPointer) ttt <- mkReg(FatPointer {meta: unpack(pack(MetaData{ag: THREAD_POINTER, tag: unpack(0)})), m: 63, b: 0, i: 0, a: 300});
        Reg#(Dunder) dunder_ttt <- mkReg(0);
        Reg#(Dover) dover_ttt <- mkReg(64);

        /***Wires***/
        Wire#(DecodedSignal) decoder_out_p0 <- mkBypassWire();
        Wire#(Atom) result_atom <- mkBypassWire();
        Wire#(RFOut) rf_out <- mkBypassWire();        

        Wire#(Atom) curr_mr_wire <- mkBypassWire();
        Wire#(ThreadState) thread_state <- mkBypassWire();
        Wire#(ErrorCode) error_code <- mkBypassWire();
        Wire#(Bool) wen <- mkBypassWire(); //write enable mem
        Wire#(FatPointer) curr_ptr_wire <- mkBypassWire();
        Wire#(CacheLine) curr_line_wire <- mkBypassWire(); //whole cache line
        Reg#(CacheLine) curr_line <- mkReg(unpack(0)); //whole cache line
        Reg#(FatPointer) curr_ptr <- mkRegU;
        Reg#(FatPointer) lcfp_ptr <- mkRegU;
        Reg#(FatPointer) new_pc <- mkRegU;
        Reg#(A) fpu_offset <- mkRegU;
        Reg#(FatPointer) decode_ptr <- mkRegU;
        Reg#(FatPointer) fhtp <- mkRegU;
        Reg#(FatPointer) ep <- mkRegU;

        /***Registers***/
        Reg#(Atom) curr_mr <- mkRegU;   // current memory read
        Reg#(Atom) curr_mw <- mkRegU;	// current memory write
        Reg#(Bit#(2)) index <- mkRegU;
        Reg#(Bool) su_block <- mkRegU;
        Reg#(FatPointer) ma <- mkRegU; //memory address
        Reg#(RF_A) itt_addr <- mkRegU;

        //control registers
        Reg#(Bool) is_lcfp <- mkRegU;
        Reg#(Bool) ptr_from_op2 <- mkRegU;
        Reg#(Bool) ptr_from_op3 <- mkRegU;
        Reg#(Bool) ptr_from_op1 <- mkRegU;
        Reg#(Bool) is_grtn <- mkRegU;
        Reg#(Bool) is_framptr <- mkRegU;
        Reg#(Bool) is_cppm <- mkRegU;
        Reg#(Bool) is_cpmp <- mkRegU;
        Reg#(Bool) is_offset_ptr <- mkRegU;
        Reg#(Bool) is_inc_ptr <- mkRegU;
        Reg#(Bool) from_base <- mkRegU;
        Reg#(Bool) is_stream <- mkRegU;
        Reg#(Bool) default_commit <- mkRegU;

        Reg#(Bool) not_insn <- mkRegU;
        
        Reg#(Bool) pc_changes <- mkRegU;
        Reg#(Bool) is_gate_insn <- mkRegU;
	Reg#(Bool) is_bracket_insn <- mkRegU;
        Reg#(A) next_pc <- mkRegU;

        //outputs of various functional units
        Reg#(TimerOut) timer_out <- mkRegU;
        Reg#(AGUOut) agu_out <- mkRegU;
        Reg#(Bool) mr_agu_out <- mkRegU;
        Reg#(DecodedSignal) decoder_out <- mkRegU;
        Reg#(ALUOut) alu_out <- mkRegU;
        Reg#(FatPUOut) fpu_out <- mkRegU;
        Reg#(TMU_R) tmu_r <- mkRegU;
        Reg#(TMU_M) tmu_m <- mkRegU;

        Wire#(Bool) tmu_allow <- mkBypassWire();        
        Wire#(TMU_R) tmu_out <- mkBypassWire();

        //simulation variables...
        Reg#(int) cycle <- mkReg(0);

        Reg#(Bit#(3)) phase <- mkReg(0);        //phase register
        //Bit#(3) phase = 2;

        Reg#(TAG_A) dc_tag <- mkReg(0);

        //bootstrap variables...
        Reg#(Bool) booting <- mkReg(False);

        //breakpoint registers
        Reg#(Bool) proc_enabled <- mkReg(False);
        Reg#(int) breakpoint_cycle <- mkRegU;
        Reg#(A) breakpoint_pc <- mkRegU;
        Reg#(Bool) break_cycle_set <- mkReg(False);
	Reg#(Bool) break_pc_set <- mkReg(False);
	
        Reg#(Tuple2#(Bit#(9),Bit#(9))) hashes <- mkRegU;
			
        /***  Rules for the processor  ***/
        /*********** Phase0 ***********/
        rule is_cycle_breakpoint_cycle (proc_enabled &&
                                        (breakpoint_cycle == 0) && break_cycle_set &&
                                        (phase == 0));
                proc_enabled <= False;
        endrule	

        rule fetch_and_decode (proc_enabled && (ts.thread_state != HALTED) &&
                               !((breakpoint_cycle == 0) && break_cycle_set) &&
                               (phase == 0));
                Instruction instr;
                Bit#(2) pc_index = truncate(pack(pcu.get_pc)) & 3;
                if (!use_ucode) begin
                        if (pc_index == 0)
                                instr = unpack(pack(imem0.a.read));
                        else if (pc_index == 1)
                                instr = unpack(pack(imem1.a.read));
                        else if (pc_index == 2)
                                instr = unpack(pack(imem2.a.read));
                        else
                                instr = unpack(pack(imem3.a.read));
                end
                else
                        instr = unpack(pack(ucode.a.read));
                
                if (!use_ucode)
                        state[valueOf(CI)] <= unpack(pack(instr));
                ci <= instr;
                //DecodedSignal tmp1 = decode(instr); //decoder.decode(instr);
		OPCODE op = unpack(truncate(pack(instr.opcode)));
                DecodedSignal dec; 

                GPR dest = instr.dest;
                GPR src1 = instr.src1;
                GPR src2 = instr.src2;
                let imm = extend(instr.imm);

	        not_insn <= instr.meta.ag != INSN;
	        dec = DecodedSignal{op: op, dest: dest, src1: src1, src2: src2, imm: imm};

                decoder_out_p0 <= dec; 
                decoder_out <= dec;
        endrule

	/*** Include code for Instruction Decoding ***/
        `include "Decoder.bsv"
        /*** _____________________________ ***/

        rule rf_read_req (proc_enabled && (ts.thread_state != HALTED) &&
                          !((breakpoint_cycle == 0) && break_cycle_set) &&
                          (phase == 0));
                op <= decoder_out_p0.op;
                ErrorCode cur_err = unpack(truncate(pack(state[valueOf(TS)]))).error_code;
                ReadIndices rf_addr;
                rf_addr[0] = decoder_out_p0.op == CPPM ? truncate(pack(decoder_out_p0.imm)) : extend(pack(decoder_out_p0.src1));
                rf_addr[1] = extend(pack(decoder_out_p0.src2));
                rf_addr[2] = extend(pack(decoder_out_p0.dest));
                rf_addr[3] = extend(pack(cur_err)) + 31;
                if (decoder_out_p0.op == RDFHTP)//get the error tp out from itt
                        itt_addr <= extend(pack(cur_err)) + 31;
                rf.read_request(rf_addr);
        endrule
	
	rule lr_mask (proc_enabled && (ts.thread_state != HALTED) &&
		      !((breakpoint_cycle == 0) && break_cycle_set) &&
		      (phase == 0));
		if (!use_ucode) begin
			if (live_mask[pack(decoder_out_p0.src1)] == 0)
				op1_is_live <= False;
			else
				op1_is_live <= True;
			if (live_mask[pack(decoder_out_p0.src2)] == 0)
				op2_is_live <= False;
			else
				op2_is_live <= True;
			if (live_mask[pack(decoder_out_p0.dest)] == 0)
				op3_is_live <= False;
			else
				op3_is_live <= True;
		end
		else begin
			op1_is_live <= True;
			op2_is_live <= True;
			op3_is_live <= True;
		end
	endrule
	
        rule update_phase0 (proc_enabled && (ts.thread_state != HALTED) && 
                            (phase == 0));
                phase <= 1;
        endrule

        /*********** Phase1 ***********/
        rule tmu_care_request (proc_enabled && (ts.thread_state != HALTED) &&
                               (phase == 1));
                tmu.care_mem_request(decoder_out.op);
        endrule        

        rule update_phase1 (proc_enabled && (phase == 1));
                phase <= 2;
        endrule
        
        /*********** Phase2 ***********/
        rule rf_read (proc_enabled && (phase == 2));
                RFOut rf_out_tmp = rf.read;
		if (!op1_is_live)
			rf_out_tmp.val0.atom = Atom{meta: MetaData{ag: ERROR, tag: public_tag}, payload: 0};
		if (!op2_is_live)
			rf_out_tmp.val1.atom = Atom{meta: MetaData{ag: ERROR, tag: public_tag}, payload: 0};
		if (!op3_is_live)
			rf_out_tmp.val2.atom = Atom{meta: MetaData{ag: ERROR, tag: public_tag}, payload: 0};
                rf_out <= rf_out_tmp;
        endrule
        
        rule update_private_from_rf (proc_enabled && (phase == 2));
                if (!use_ucode) begin
			state[valueOf(OP1)] <= rf_out.val0.atom;
                        state[valueOf(OP2)] <= rf_out.val1.atom;
                        state[valueOf(OP3)] <= rf_out.val2.atom;
                end
        endrule
        
        rule set_ops_from_rf (proc_enabled && (phase == 2));
                op1 <= rf_out.val0.atom;
                op2 <= rf_out.val1.atom;
                op3 <= rf_out.val2.atom;
                op1_dunder <= rf_out.val0.dunder;
                op1_dover <= rf_out.val0.dover;
                fhtp <= unpack(pack(rf_out.val3.atom));        
        endrule
        
        rule set_mem_write (proc_enabled && (phase == 2));
                Bit#(6) state_idx = truncate(decoder_out.imm) - 48;
                if (is_cppm) begin
                        if (decoder_out.imm < 48)
                                curr_mw <= rf_out.val0.atom;
                        else
                                curr_mw <= state[state_idx];
                end
                else
                        curr_mw <= rf_out.val0.atom;
        endrule

        rule tmu_care_lookup (proc_enabled && (phase == 2));
                tmu.care_mem_lookup;
        endrule

        rule tmu_phase1 (proc_enabled && (phase == 2));
                tmu.tmu_rc_req(unpack(truncate(pack(rf_out.val0.atom.payload))));
        endrule

        rule pick_pointer (proc_enabled && (phase == 2));
                FatPointer ptr = FatPointer{meta: MetaData{ag: unpack(0), tag: unpack(0)}, m: 1, b: 0, i: 0, a: 0};
                Dunder tmp_dunder = ?;
                Dover tmp_dover = ?;
                if (is_lcfp) begin//lcfp
                        ptr = pc;
                        tmp_dunder = pcu.get_pc_dunder;
                        tmp_dover = pcu.get_pc_dover;
                end
                else if (ptr_from_op1 || (decoder_out_p0.op == MVRR) || (decoder_out_p0.op == CPRR)) begin
                        ptr = unpack(pack(rf_out.val0.atom));	//includes strb, stry
                        tmp_dunder = rf_out.val0.dunder;
                        tmp_dover = rf_out.val0.dover;
                end
                else if (ptr_from_op2 || (decoder_out.op == REGRP)) begin //{xxRM,stwy}, offp, stwb
                        ptr = unpack(pack(rf_out.val1.atom));
                        tmp_dunder = rf_out.val1.dunder;
                        tmp_dover = rf_out.val1.dover;
                end
                else if (ptr_from_op3) begin
                        ptr = unpack(pack(rf_out.val2.atom));
                        tmp_dunder = rf_out.val2.dunder;
                        tmp_dover = rf_out.val2.dover;
                end
                else if (is_grtn) begin
                        let gs_ptr = gate_stack.get_gsp;
                        ptr.meta = gs_ptr.meta;
                        ptr.i = gs_ptr.i;
                        ptr.m = gs_ptr.m;
                        ptr.b = gs_ptr.b;
                        ptr.a = gs_ptr.a-4;
                        tmp_dunder = gate_stack.get_gsp_dunder;
                        tmp_dover = gate_stack.get_gsp_dover;
                end
                else if (is_cpmp || is_cppm) begin
                        ptr = tmp_tp;
                        tmp_dunder = dunder_tmp_tp;
                        tmp_dover = dover_tmp_tp;
                end

                A tmp_offset;
                if (is_offset_ptr)
                        tmp_offset = truncate(rf_out.val0.atom.payload);
                else if (is_lcfp)
                        tmp_offset = extend(decoder_out.imm);
                else if (is_inc_ptr)
                        tmp_offset = unpack(1);
                else
                        tmp_offset = unpack(0);
                dunder <= tmp_dunder;
                dover <= tmp_dover;
                Dunder base_offset = from_base ? unpack(tmp_offset) - tmp_dunder : unpack(tmp_offset);
                fpu_offset <= pack(base_offset);
                curr_ptr_wire <= ptr;
                curr_ptr <= ptr;
                index <= truncate(ptr.a) & 3;
        endrule
		
        //perform an address match inside the Write Buffer..
        rule wb_lookup ((proc_enabled && in_trans) && (phase == 2));
                write_buffer.addr_match(truncate(curr_ptr_wire.a));
        endrule

        rule take_branch (proc_enabled && (phase == 2));
                bu.pick_offset(op, zeroExtend(decoder_out.imm));
        endrule
        
        rule update_phase2 (phase == 2);
                phase <= 3;
        endrule

        /*********** Phase3 ***********/

        //load happens in method

        rule fat_pointer_check (proc_enabled && (phase == 3));
                let tmp_fpuout = fpu.update_pointer(curr_ptr, dunder, dover, fpu_offset, is_stream);
                fpu_out <= tuple2((is_lcfp || ptr_from_op2 || ptr_from_op3 || ptr_from_op1 || is_grtn || is_inc_ptr) && (curr_ptr.meta.ag == ERROR), tpl_1(tmp_fpuout));
                dunder <= tpl_2(tmp_fpuout);
                dover <= tpl_3(tmp_fpuout);
                ma <= tpl_1(tmp_fpuout);
        endrule

        rule alu_exec (proc_enabled && (phase == 3));
                ALUOut tmp = alu.do_alu(op, op1.payload, op2.payload);
                alu_out <= tmp;
        endrule
	
	rule make_framptr (proc_enabled && (phase == 3) && !ren && is_framptr);
		FatPointer fp2 = unpack(pack(op2));
		B new_b = pack(op1)[11:6];
		L new_l = pack(op1)[5:0];
		
		I new_i = truncate(fp2.a >> new_b);
		M new_m = truncate(truncate(fp2.a >> new_b) + new_l);
		
		FatPointer ptr = ?;
		ptr.i = new_i;
		ptr.b = new_b;
		ptr.m = new_m;
		ptr.a = fp2.a;
		
		curr_mr_wire <= unpack(pack(ptr));
	endrule


        rule atomic_group_check (proc_enabled && (phase == 3));
                AtomicGroup src1_ag = op1.meta.ag;
                //cppm and cpmp needs to make sure that tp is thread pointer
                AtomicGroup src2_ag = is_cppm || is_cpmp ? tp.meta.ag : op2.meta.ag;
                AtomicGroup src3_ag = op3.meta.ag;
                let agu_res = agu.do_ag_check(op, src1_ag, src2_ag, src3_ag);
                agu_out <= agu_res;
        endrule

        rule dummy_read ((phase == 3) && !ren && !is_framptr);
                curr_mr_wire <= unpack(0);
        endrule

        rule tmu_mask_dc (proc_enabled && (phase == 3));
                let wb_res <- write_buffer.do_read;
                Atom tmu_memr;
                if (in_trans && tpl_1(wb_res))
                        tmu_memr = tpl_2(wb_res);
                else if (is_gate_insn)
                        tmu_memr = curr_line_wire.a2;
                else
                        tmu_memr = curr_mr_wire;
                let hash_vals <- tmu.mask_dc(tp.meta.tag, ci.meta.tag,
                                             pcu.get_pc.meta.tag, 
                                             use_ucode ? unpack(truncate(pack(proc_auth))) : unpack(truncate(pack(auth))),
                                             op1.meta.tag, op2.meta.tag,
                                             op3.meta.tag, tmu_memr.meta.tag);
                curr_mr <= tmu_memr;
                hashes <= hash_vals;
        endrule

        rule bu_resolve (proc_enabled && (phase == 3));
                Int#(PAYLOAD_WIDTH) int_payload = unpack(op1.payload);
                Bool update_from_branch = ((decoder_out.op == JMP) ||
                                           (((decoder_out.op == STWB) || (decoder_out.op == STRB)) && su_block) ||
                                           ((decoder_out.op == BEQ) && (op1.payload == 0)) ||
                                           ((decoder_out.op == BNE) && (op1.payload != 0)) ||
                                           ((decoder_out.op == BPOS) && (int_payload > 0)) || 
					   ((decoder_out.op == BNEG) && (int_payload < 0)));
                pc_from_base <= update_from_branch;
                if (decoder_out.op == ENDU)
                        next_pc <= pc.a;
                else if (pc_changes) begin
                        FatPointer pc_from_gate = unpack(pack(curr_line_wire.a0));
                        FatPointer pc_from_call = unpack(pack(op1));
                        next_pc <= is_gate_insn ? pc_from_gate.a : pc_from_call.a;
                end
                else if (update_from_branch) begin
                        next_pc <= pcu.get_pc.a - pack(pcu.get_pc_dunder) + bu.get_offset;
                end
                else
                        next_pc <= truncate(pcu.get_pc.a) + 1;
                bu.set_offset(update_from_branch ? bu.get_offset : 1);
        endrule
        
        rule select_pc_changing (proc_enabled && pc_changes && (phase == 3));
                if (is_gate_insn)
                        new_pc <= unpack(pack(curr_line_wire.a0));
                else if (decoder_out.op == ENDU)
                        new_pc <= pc;
                else
                        new_pc <= unpack(pack(op1));
        endrule
        
        rule select_pc_decode (proc_enabled && (phase == 3));
                if (is_gate_insn) //update env ptr
                        decode_ptr <= unpack(pack(curr_line_wire.a2));
                else if (decoder_out.op == RDFHTP)
                        decode_ptr <= fhtp;
                else if (decoder_out.op == CPRR)
                        decode_ptr <= unpack(pack(op1));
                else
                        decode_ptr <= unpack(pack(curr_mr_wire));
        endrule

        rule update_phase3 (phase == 3);
                phase <= 4;
        endrule


        /********* Phase 4 ************/
	
        rule tmu_g_table_lookup (proc_enabled && (phase == 4) && (curr_mr.meta.ag != FWD_POINTER));
                tmu.g_table_lookup;
        endrule
        
        rule update_hash_vals (proc_enabled && (phase == 4) && (curr_mr.meta.ag != FWD_POINTER));
                //TODO:need to actually get right tags
                if (!use_ucode) begin
                        state[valueOf(H1)] <= unpack(extend(tpl_1(hashes)));
                        state[valueOf(H2)] <= unpack(extend(tpl_2(hashes)));
                end        
        endrule

        rule mr_ag_check(proc_enabled && (phase == 4) && (curr_mr.meta.ag != FWD_POINTER));
                mr_agu_out <= agu.do_mr_ag_check(op, curr_line.a0.meta.ag, curr_line.a1.meta.ag, curr_line.a2.meta.ag);
        endrule

        rule pcu_compute (proc_enabled && (phase == 4) && (curr_mr.meta.ag != FWD_POINTER));
                pcu.compute_dist(new_pc);
        endrule

        rule fat_pointer_check_bound (proc_enabled && !is_lcfp && (phase == 4) && (curr_mr.meta.ag != FWD_POINTER));
                let fpu_res = fpu.decode(decode_ptr);                
                dunder_mem <= tpl_1(fpu_res);
                dover_mem <= tpl_2(fpu_res);
        endrule
        
        rule fat_pointer_decode_lcfp (proc_enabled && is_lcfp && (phase == 4) && (curr_mr.meta.ag != FWD_POINTER));
                M new_m = tpl_2(fpu_out).a[5:0];
                I new_i = tpl_2(fpu_out).a[5:0];
                lcfp_ptr <= FatPointer{meta: MetaData{ag: tpl_2(agu_out),
                                                       tag: unpack(0)},
                                        m: new_m, b: unpack(0), i: new_i, a: ma.a};
                dunder_mem <= unpack(0);
                dover_mem <= unpack(0);
        endrule

        rule update_timer (proc_enabled && (phase == 4) && (curr_mr.meta.ag != FWD_POINTER));
                let tmp_timer_out <- timer.do_timer(op, unpack(op1.payload));
                timer_out <= tmp_timer_out;
        endrule

        rule su_res (proc_enabled && (phase == 4) && (curr_mr.meta.ag != FWD_POINTER));
                su_block <= su.do_stream(op, op1.meta.ag, op2.meta.ag, curr_mr.meta.ag);
        endrule

        rule update_phase4 (proc_enabled && (phase == 4) && (curr_mr.meta.ag != FWD_POINTER));
                phase <= 5;
        endrule
        
        
        rule is_fwd_ptr (proc_enabled && (phase == 4) && (curr_mr.meta.ag == FWD_POINTER));
		Atom atom = curr_mr;
		GPR new_dest = ?;
		if (ptr_from_op2) begin
			atom.meta.ag = op2.meta.ag;
			new_dest = decoder_out.src2;
		end
		else if (ptr_from_op3) begin	
			atom.meta.ag = op3.meta.ag;
			new_dest = decoder_out.dest;
		end
		else begin
			atom.meta.ag = op1.meta.ag;
			new_dest = decoder_out.src1;
		end

		RegAtom writeback = RegAtom{atom: atom, dunder: dunder_mem, dover: dover_mem};
		rf.write_port1(extend(pack(new_dest)), writeback);
		phase <= 0;
	endrule


        /********* Phase 5 ***********/
	
	
        rule tmu_mr_lookup (proc_enabled && (phase == 5));
                tmu.mr_lookup;
        endrule

        rule update_phase5 (proc_enabled && (phase == 5) && (curr_mr.meta.ag != FWD_POINTER));
                phase <= 6;
        endrule
        
        /********* Phase 6 ***********/

        rule tmu_mr_match (proc_enabled && (phase == 6));
                tmu_allow <= tpl_1(tmu.mr_match);
                tmu_out <= tpl_2(tmu.mr_match);
        endrule
        
        /*** include Result Concatenate code ***/
        `include "ResultCat.bsv"
        /*** _______________________________ ***/

        rule update_thread_status (proc_enabled && (phase == 6));
                //update error code
                let is_timer_error = tpl_1(timer_out);
                let is_frame_bounds_error = (tpl_1(fpu_out) || !pcu.in_bounds);
                let is_ag_error = (tpl_1(agu_out) || pcu.pc_ag_check || mr_agu_out || not_insn);
                let error_num;
                if (op == HALT) begin
                        error_num = NO_ERROR;
                        error_code <= NO_ERROR;
                        thread_state <= HALTED;
                end
                // else if (is_timer_error) begin
                //         error_num = TIMER_INTERRUPT;
                //         error_code <= TIMER_INTERRUPT;
                //         thread_state <= ts.thread_state;
                // end
                else if (is_frame_bounds_error) begin
                        error_num = FRAME_BOUNDS_ERROR;
                        error_code <= FRAME_BOUNDS_ERROR;
                        thread_state <= FAULTED;
                end
                else if (is_ag_error) begin
                        error_num = AG_ERROR;
                        error_code <= AG_ERROR;
                        thread_state <= FAULTED;
                end
                else if (su_block) begin
                        error_num = STREAM_BLOCKING;
                        error_code <= STREAM_BLOCKING;
                        thread_state <= FAULTED;
                end
                else begin
                        error_num = NO_ERROR;
                        error_code <= NO_ERROR;
                        thread_state <= ts.thread_state;
                end
        endrule


        /*** Include code for Commit Stage ***/
        `include "Commit.bsv"
        /*** _____________________________ ***/

        /*** Include all the method definitions ***/
        `include "Methods.bsv"
        /*** __________________________________ ***/

endmodule: mkSAFE
