/************************************************************************************************
 * Copyright  (c) 2011, 2012 University of Pennsylvania
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose, without fee, and without a
 * written agreement is hereby granted, provided that the above copyright
 * notice and this paragraph and the following two paragraphs appear in
 * all copies.
 *
 * IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
 * HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
 * AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 ***********************************************************************************************/

/****************************************************
Perfect Hash TMU for NoCopy LabelModel
Author: Udit Dhawan
IC Group, University of Pennsylvania

Notes:
The TMU performs table reads in both static ground table and the dynamic tables and 
takes the one that matches...if none match, then it faults to the TMU manager which 
should compute the tags and write them back to the dynamic tables.
*****************************************************/

import SAFETypes::*;
import Hash::*;
import BRAM::*;
import BRAMCore::*;

interface TMUIfc;
	method Action care_mem_request(OPCODE op);
	method Action care_mem_lookup();
	method ActionValue#(Tuple2#(Bit#(9),Bit#(9))) mask_dc(Tag tptag, Tag citag, Tag pctag, Tag authority, Tag op1tag, Tag op2tag, Tag op3tag, Tag mrtag);
	method Action mr_lookup();
        method Tuple2#(Bool,TMU_R) mr_match();
	method Action g_table_lookup();
	method Action tmu_unload(Bit#(8) slot);
	method Action tmu_load(Bit#(9) slot, Bit#(5) field, TAG_A tag);
	method Action tmu_rc_req(Bit#(8) slot);
	method Bit#(16) tmu_rc_read();
	method Action tmu_rc_update(Bit#(16) new_count);
	method Action set_dc_tag(TAG_A new_dc_tag);
        method Action care_mem_load(Bit#(8) addr, Atom val);
        method Action res_care_load(Bit#(8) addr, Atom val);
        method Action gnd_g1_load(Bit#(9) addr, Atom val);
        method Action gnd_g2_load(Bit#(9) addr, Atom val);
        method Action gnd_mr_load(Bit#(8) addr, TMU_MR val);
endinterface: TMUIfc

(* synthesize *)
module mkTMU(TMUIfc);

	BRAM_DUAL_PORT#(Bit#(8), Atom) care_mem <- mkBRAMCore2Load(128, False, "care.img", False);
	BRAM_DUAL_PORT#(Bit#(8), Atom) res_care_mem <- mkBRAMCore2Load(128, False, "res_care.img", False);

	//G tables for static TMU
	BRAM_DUAL_PORT#(Bit#(9), Atom) gnd_g1_table <- mkBRAMCore2Load(512, False, "g.img", False);
	BRAM_DUAL_PORT#(Bit#(9), Atom) gnd_g2_table <- mkBRAMCore2Load(512, False, "g.img", False);

	//G Tables for dynamic TMU
	BRAM_DUAL_PORT#(Bit#(9), Atom) dyn_g1_table <- mkBRAMCore2(512, False);
	BRAM_DUAL_PORT#(Bit#(9), Atom) dyn_g2_table <- mkBRAMCore2(512, False);

	// BRAMs for static ground rules table...
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_m_tptag <- mkBRAMCore2Load(256, False, "mr0.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_m_citag <- mkBRAMCore2Load(256, False, "mr1.img", False);
	BRAM_DUAL_PORT#(Bit#(8), Bit#(8)) gnd_m_ciop <- mkBRAMCore2Load(256, False, "mr2.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_m_pctag <- mkBRAMCore2Load(256, False, "mr3.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_m_auth <- mkBRAMCore2Load(256, False, "mr4.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_m_op1tag <- mkBRAMCore2Load(256, False, "mr5.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_m_op2tag <- mkBRAMCore2Load(256, False, "mr6.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_m_op3tag <- mkBRAMCore2Load(256, False, "mr7.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_m_mrtag <- mkBRAMCore2Load(256, False, "mr8.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_r_op3res <- mkBRAMCore2Load(256, False, "mr9.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_r_op2res <- mkBRAMCore2Load(256, False, "mr10.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_r_op1res <- mkBRAMCore2Load(256, False, "mr11.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_r_mwres <- mkBRAMCore2Load(256, False, "mr12.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_r_pcres <- mkBRAMCore2Load(256, False, "mr13.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_r_tpres <- mkBRAMCore2Load(256, False, "mr14.img", False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) gnd_r_auth <- mkBRAMCore2Load(256, False, "mr15.img", False);
	// skipping monitor field for now	--- 01/02/2012

	// BRAMs for dynamic tables
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_m_tptag <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_m_citag <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), Bit#(8)) dyn_m_ciop <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_m_pctag <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_m_auth <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_m_op1tag <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_m_op2tag <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_m_op3tag <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_m_mrtag <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_r_op3res <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_r_op2res <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_r_op1res <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_r_mwres <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_r_pcres <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_r_tpres <- mkBRAMCore2(256, False);
	BRAM_DUAL_PORT#(Bit#(8), TAG_A) dyn_r_auth <- mkBRAMCore2(256, False);
	// skipping monitor field for now	--- 01/02/2012

	// hit counters for the dynamic TMU
	BRAM_DUAL_PORT#(Bit#(8), Bit#(16)) hit_counters <- mkBRAMCore2(256, False);

	HashIfc h1 <- mkHash(9,2,7,1,6,1,0,8,4,2);
	HashIfc h2 <- mkHash(9,4,1,5,9,5,2,9,2,1);

	Reg#(Bit#(8)) opgrp <- mkRegU;
	Reg#(Bit#(9)) care_bits <- mkRegU;
	Reg#(Bit#(9)) res_care_bits <- mkRegU;
	Reg#(TMU_M) tmu_m <- mkRegU;
        Reg#(TMU_M) tmu_m_origin <- mkRegU;
	Reg#(Bit#(8)) gnd_mr_addr <- mkRegU;
	Reg#(Bit#(8)) dyn_mr_addr <- mkRegU;
        
        Reg#(TMU_MR) gnd_mr <- mkRegU;
        Reg#(TMU_MR) dyn_mr <- mkRegU;        

	Reg#(TAG_A) dc_tag <- mkRegU;


	function Bit#(8) opcode_to_opgrp(OPCODE op);
		Bit#(8) opgrp=0;

		/* Logic for opcode to opgroup conversion... */
		if ((op == ADD) || (op == SUB) || (op == AND) || (op == OR) || (op == SHR) || (op == SHL) || (op == MUL) || (op == TEST))
			opgrp = 'h1;
		else if (op == LCFP)
			opgrp = 'h2;
		else if ((op == CPMR))
			opgrp = 'h3;
		else if ((op == CPRM))
			opgrp = 'h4;
		else if ((op == YIELD))
			opgrp = 'h5;
		else if ((op == HALT))
			opgrp = 'h6;
		else if ((op == BEQ) || (op == BNE) || (op == BNEG) || (op == BPOS))
			opgrp = 'h7;
		else if ((op == GCALL) || (op == GJMP))
			opgrp = 'h8;
		else if ((op == GRTN))
			opgrp = 'h9;
		else if ((op == RUNT))
			opgrp = 'ha;
		else if ((op == GTIME))
			opgrp = 'hb;
		else if ((op == REMT))
			opgrp = 'hc;
		else if ((op == RDTIME))
			opgrp = 'hd;
		else if ((op == STRB))
			opgrp = 'he;
		else if ((op == STRY))
			opgrp = 'hf;
		else if ((op == STWB))
			opgrp = 'h10;
		else if ((op == STWY))
			opgrp = 'h11;
		else if ((op == CPRR))
			opgrp = 'h12;
		else if ((op == MVRR))
			opgrp = 'h13;
		else if ((op == FRAMPTR))
			opgrp = 'h14;
		else if ((op == OFFP))
			opgrp = 'h15;
		else if ((op == OFFLP))
			opgrp = 'h16;
		else if ((op == TMUL))
			opgrp = 'h17;
		else if ((op == TMUU))
			opgrp = 'h18;
		else if ((op == RETAG))
			opgrp = 'h19;
		else if ((op == INTAG))
			opgrp = 'h1a;
		else if ((op == MAKEP))
			opgrp = 'h1b;
		else if ((op == NEWT))
			opgrp = 'h1c;
		else if ((op == INP))
			opgrp = 'h1d;
		else if ((op == SETA))
			opgrp = 'h1e;
		else if ((op == REGRP))
			opgrp = 'h1f;
		else if ((op == MVRM))
			opgrp = 'h20;
		else if ((op == MVMR))
			opgrp = 'h21;
		else if ((op == RESUMET))
			opgrp = 'h22;
		else if ((op == GATE))
			opgrp = 'h24;
		else if ((op == FJMP))
			opgrp = 'h25;
		else if ((op == NEWG))
			opgrp = 'h26;
		else if ((op == TOTAG))
			opgrp = 'h27;
		else if ((op == RFLCT))
			opgrp = 'h28;
		else if ((op == GFWRITE))
			opgrp = 'h2a;
		else if ((op == FPHASH))
			opgrp = 'h2b;
		else if ((op == RFLCTP))
			opgrp = 'h29;
		else if ((op == RETAGPC))
			opgrp = 'h2c;
		else if ((op == SETTAG))
			opgrp = 'h2d;
		else if ((op == CALL))
			opgrp = 'h2e;
		else if ((op == MAKET))
			opgrp = 'h2f;
		else if ((op == INGRP))
			opgrp = 'h30;
		else if ((op == TAGOF))
			opgrp = 'h31;
		else if ((op == TMURC))
			opgrp = 'h32;
		else if ((op == INA))
			opgrp = 'h33;
		else if ((op == CPMT))
			opgrp = 'h34;
		else if ((op == CPIO))
			opgrp = 'h35;
		else if ((op == CPAR))
			opgrp = 'h36;
		else if ((op == CPHR))
			opgrp = 'h37;
		else if ((op == OFFTP))
			opgrp = 'h38;
		else if ((op == JMP))
			opgrp = 'h39;
		else if ((op == TRANSBEGIN) || (op == TRANSEND))
			opgrp = 'h3a;
		else if ((op == TESTGRP))
			opgrp = 'h3b;
		else if ((op == GACALL))
			opgrp = 'h3c;
		else if ((op == GAJMP))
			opgrp = 'h3d;
		else if ((op == CPPM) || (op == CPMP) || (op == ENDU) || (op == RDFHTP) || (op == SETFHTP))
			opgrp = 'h3e;
		else
			opgrp = 'h0;

		return opgrp;
	endfunction


	function TMU_MR modify_mr(TMU_MR mr, Bit#(5) field, TAG_A new_tag);
		//create a mask for masking out the field to be replaced...
		TMU_MR new_mr = mr;

		case(field)
			0: new_mr.match_vector.tptag  = new_tag;
			1: new_mr.match_vector.citag  = new_tag;
			2: new_mr.match_vector.opgrp  = unpack(truncate(pack(new_tag)));
			3: new_mr.match_vector.pctag  = new_tag;
			4: new_mr.match_vector.auth = new_tag;
			5: new_mr.match_vector.op1tag  = new_tag;
			6: new_mr.match_vector.op2tag  = new_tag;
			7: new_mr.match_vector.op3tag  = new_tag;
			8: new_mr.match_vector.mrtag  = new_tag;
			9: new_mr.res_vector.op3res  = new_tag;
			10: new_mr.res_vector.op2res  = new_tag;
			11: new_mr.res_vector.op1res  = new_tag;
			12: new_mr.res_vector.pcres  = new_tag;
			13: new_mr.res_vector.mwres  = new_tag;
			14: new_mr.res_vector.tpres  = new_tag;
			15: new_mr.res_vector.auth  = new_tag;
		endcase

		return new_mr;
	endfunction


	method Action care_mem_request(OPCODE op);	//put read request in phase 0
		let tmp = opcode_to_opgrp(op);
		care_mem.a.put(False, tmp, ?);
		res_care_mem.a.put(False, tmp, ?);
		opgrp <= tmp;
	endmethod


	method Action care_mem_lookup();	//read in phase 1
		Bit#(9) tmp0 = truncate(pack(care_mem.a.read));
		Bit#(9) tmp1 = truncate(pack(res_care_mem.a.read));

		care_bits <= tmp0;
		res_care_bits <= tmp1;
	endmethod


	method ActionValue#(Tuple2#(Bit#(9),Bit#(9))) mask_dc(Tag tptag, Tag citag, Tag pctag, Tag authority, Tag op1tag, Tag op2tag, Tag op3tag, Tag mrtag);	//in phase 2, after curr_mr_wire has been written in phase 2...

		TMU_M temp = unpack(0);
                TMU_M origin = unpack(0);
                origin.mrtag = truncate(mrtag.a);
                origin.op3tag = truncate(op3tag.a);
                origin.op2tag = truncate(op2tag.a);
		origin.op1tag = truncate(op1tag.a);
		origin.auth = truncate(authority.a);
		origin.pctag = truncate(pctag.a);
		origin.citag = truncate(citag.a);
		origin.tptag = truncate(tptag.a);


		if((care_bits&1)==1) begin
			temp.mrtag = truncate(mrtag.a);
		end
		else begin
			temp.mrtag = dc_tag;
		end
		if(((care_bits>>1)&1)==1) begin
			temp.op3tag = truncate(op3tag.a);
		end
		else begin
			temp.op3tag = dc_tag;
		end
		if(((care_bits>>2)&1)==1) begin
			temp.op2tag = truncate(op2tag.a);
		end
		else begin
			temp.op2tag = dc_tag;
		end
		if(((care_bits>>3)&1)==1) begin
			temp.op1tag = truncate(op1tag.a);
		end
		else begin
			temp.op1tag = dc_tag;
		end
		if(((care_bits>>4)&1)==1) begin
			temp.auth = truncate(authority.a);
		end
		else begin
			temp.auth = dc_tag;
		end
		if(((care_bits>>5)&1)==1) begin
			temp.pctag = truncate(pctag.a);
		end
		else begin
			temp.pctag = dc_tag;
		end
		//we always care about the opgroup
//		if(((care_bits>>6)&1)==1) begin
//			temp.opgrp = opgrp;
//		end
		if(((care_bits>>6)&1)==1) begin
			temp.citag = truncate(citag.a);
		end
		else begin
			temp.citag = dc_tag;
		end
		if(((care_bits>>7)&1)==1) begin
			temp.tptag = truncate(tptag.a);
		end
		else begin
			temp.tptag = dc_tag;
		end

		temp.opgrp = opgrp;
        
		/*  we have M vector in temp now, compute the hash values for the same.... */
		Bit#(9) hash_val1 = h1.compute_hash(temp);
		Bit#(9) hash_val2 = h2.compute_hash(temp);

		/* put the read request for G tables */
		gnd_g1_table.a.put(False, hash_val1, ?);
		dyn_g1_table.a.put(False, hash_val1, ?);
		gnd_g2_table.a.put(False, hash_val2, ?);
		dyn_g2_table.a.put(False, hash_val2, ?);

		tmu_m <= temp;
                tmu_m_origin <= origin;

		return tuple2(hash_val1,hash_val2);
	endmethod


	method Action g_table_lookup();		///read the G tables in phase 3
		Bit#(8) gnd_g1 = truncate(pack(gnd_g1_table.a.read));
		Bit#(8) gnd_g2 = truncate(pack(gnd_g2_table.a.read));
		Bit#(8) dyn_g1 = truncate(pack(dyn_g1_table.a.read));
		Bit#(8) dyn_g2 = truncate(pack(dyn_g2_table.a.read));

		Bit#(8) gnd_mr_addr_tmp = gnd_g1 ^ gnd_g2;
		Bit#(8) dyn_mr_addr_tmp = dyn_g1 ^ dyn_g2;

		gnd_mr_addr <= gnd_mr_addr_tmp;
		dyn_mr_addr <= dyn_mr_addr_tmp;

		/* put the read request for all the MR field tables */
		gnd_m_tptag.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_m_tptag.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_m_citag.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_m_citag.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_m_ciop.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_m_ciop.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_m_pctag.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_m_pctag.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_m_auth.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_m_auth.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_m_op1tag.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_m_op1tag.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_m_op2tag.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_m_op2tag.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_m_op3tag.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_m_op3tag.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_m_mrtag.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_m_mrtag.a.put(False, dyn_mr_addr_tmp, ?);

		gnd_r_op3res.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_r_op3res.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_r_op2res.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_r_op2res.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_r_op1res.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_r_op1res.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_r_mwres.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_r_mwres.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_r_pcres.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_r_pcres.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_r_tpres.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_r_tpres.a.put(False, dyn_mr_addr_tmp, ?);
		gnd_r_auth.a.put(False, gnd_mr_addr_tmp, ?);
		dyn_r_auth.a.put(False, dyn_mr_addr_tmp, ?);

	endmethod

	method Action mr_lookup();   //read the result tags in phase 4

		// read the M tables
		TAG_A gnd_m_0 = gnd_m_tptag.a.read;
		TAG_A dyn_m_0 = dyn_m_tptag.a.read;
		TAG_A gnd_m_1 = gnd_m_citag.a.read;
		TAG_A dyn_m_1= 	dyn_m_citag.a.read;
		Bit#(8) gnd_m_2 = gnd_m_ciop.a.read;
		Bit#(8) dyn_m_2 = dyn_m_ciop.a.read;
		TAG_A gnd_m_3 =	gnd_m_pctag.a.read;
		TAG_A dyn_m_3 =	dyn_m_pctag.a.read;
		TAG_A gnd_m_4 =	gnd_m_auth.a.read;
		TAG_A dyn_m_4 =	dyn_m_auth.a.read;
		TAG_A gnd_m_5 =	gnd_m_op1tag.a.read;
		TAG_A dyn_m_5 =	dyn_m_op1tag.a.read;
		TAG_A gnd_m_6 =	gnd_m_op2tag.a.read;
		TAG_A dyn_m_6 =	dyn_m_op2tag.a.read;
		TAG_A gnd_m_7 =	gnd_m_op3tag.a.read;
		TAG_A dyn_m_7 =	dyn_m_op3tag.a.read;
		TAG_A gnd_m_8 =	gnd_m_mrtag.a.read;
		TAG_A dyn_m_8 =	dyn_m_mrtag.a.read;

		// read the R tables
		TAG_A gnd_r_0 = gnd_r_op3res.a.read;
		TAG_A dyn_r_0 = dyn_r_op3res.a.read;
		TAG_A gnd_r_1 = gnd_r_op2res.a.read;
		TAG_A dyn_r_1 = dyn_r_op2res.a.read;
		TAG_A gnd_r_2 =	gnd_r_op1res.a.read;
		TAG_A dyn_r_2 =	dyn_r_op1res.a.read;
		TAG_A gnd_r_3 =	gnd_r_mwres.a.read;
		TAG_A dyn_r_3 =	dyn_r_mwres.a.read;
		TAG_A gnd_r_4 =	gnd_r_pcres.a.read;
		TAG_A dyn_r_4 =	dyn_r_pcres.a.read;
		TAG_A gnd_r_5 =	gnd_r_tpres.a.read;
		TAG_A dyn_r_5 =	dyn_r_tpres.a.read;
		TAG_A gnd_r_6 =	gnd_r_auth.a.read;
		TAG_A dyn_r_6 =	dyn_r_auth.a.read;


		gnd_mr <= TMU_MR{match_vector: TMU_M{tptag: gnd_m_0, citag: gnd_m_1, opgrp: gnd_m_2, pctag: gnd_m_3, auth: gnd_m_4, op1tag: gnd_m_5, op2tag: gnd_m_6, op3tag: gnd_m_7, mrtag: gnd_m_8}, res_vector: TMU_R{op3res: gnd_r_0, op2res: gnd_r_1, op1res: gnd_r_2, mwres: gnd_r_3, pcres: gnd_r_4, tpres: gnd_r_5, auth: gnd_r_6}};
		dyn_mr <= TMU_MR{match_vector: TMU_M{tptag: dyn_m_0, citag: dyn_m_1, opgrp: dyn_m_2, pctag: dyn_m_3, auth: dyn_m_4, op1tag: dyn_m_5, op2tag: dyn_m_6, op3tag: dyn_m_7, mrtag: dyn_m_8}, res_vector: TMU_R{op3res: dyn_r_0, op2res: dyn_r_1, op1res: dyn_r_2, mwres: dyn_r_3, pcres: dyn_r_4, tpres: dyn_r_5, auth: dyn_r_6}};
                
	endmethod
        
        method Tuple2#(Bool,TMU_R) mr_match();
        	TMU_R tmu_r = unpack(0);
		tmu_r.auth = tmu_m_origin.auth;
		tmu_r.tpres = tmu_m_origin.tptag;
		tmu_r.pcres = tmu_m_origin.pctag;
		tmu_r.mwres = tmu_m_origin.mrtag;
		tmu_r.op1res = tmu_m_origin.op1tag;
		tmu_r.op2res = tmu_m_origin.op2tag;
		tmu_r.op3res = tmu_m_origin.op3tag;

		Bool gnd_tmu_allow = False;
		Bool dyn_tmu_allow = False;

		//see if the input M vector matches from the MR output or not...
		if(((tmu_m.tptag==gnd_mr.match_vector.tptag) || (((care_bits>>7)&1)==0))
			&& ((tmu_m.citag==gnd_mr.match_vector.citag) || (((care_bits>>6)&1)==0))
			&& ((tmu_m.opgrp==gnd_mr.match_vector.opgrp))
			&& ((tmu_m.pctag==gnd_mr.match_vector.pctag) || (((care_bits>>5)&1)==0))
			&& ((tmu_m.auth==gnd_mr.match_vector.auth) || (((care_bits>>4)&1)==0))
			&& ((tmu_m.op1tag==gnd_mr.match_vector.op1tag) || (((care_bits>>3)&1)==0))
			&& ((tmu_m.op2tag==gnd_mr.match_vector.op2tag) || (((care_bits>>2)&1)==0))
			&& ((tmu_m.op3tag==gnd_mr.match_vector.op3tag) || (((care_bits>>1)&1)==0))
			&& ((tmu_m.mrtag==gnd_mr.match_vector.mrtag) || ((care_bits&1)==0)) )
			begin
				gnd_tmu_allow = True;
				tmu_r = gnd_mr.res_vector;
			end
		else if(((tmu_m.tptag==dyn_mr.match_vector.tptag) || (((care_bits>>7)&1)==0))
			&& ((tmu_m.citag==dyn_mr.match_vector.citag) || (((care_bits>>6)&1)==0))
			&& ((tmu_m.opgrp==dyn_mr.match_vector.opgrp))
			&& ((tmu_m.pctag==dyn_mr.match_vector.pctag) || (((care_bits>>5)&1)==0))
			&& ((tmu_m.auth==dyn_mr.match_vector.auth) || (((care_bits>>4)&1)==0))
			&& ((tmu_m.op1tag==dyn_mr.match_vector.op1tag) || (((care_bits>>3)&1)==0))
			&& ((tmu_m.op2tag==dyn_mr.match_vector.op2tag) || (((care_bits>>2)&1)==0))
			&& ((tmu_m.op3tag==dyn_mr.match_vector.op3tag) || (((care_bits>>1)&1)==0))
			&& ((tmu_m.mrtag==dyn_mr.match_vector.mrtag) || ((care_bits&1)==0)) )
			begin
				dyn_tmu_allow = True;
				tmu_r = dyn_mr.res_vector;
			end

		Bool tmu_allow = gnd_tmu_allow || dyn_tmu_allow;

		//mask out dont care fields in the result vector...
		if ((res_care_bits & 1) == 0)
			tmu_r.auth = tmu_m_origin.auth;
		if (((res_care_bits >> 1) & 1) == 0)
			tmu_r.tpres = tmu_m_origin.tptag;
		if (((res_care_bits >> 2) & 1) == 0) 
			tmu_r.pcres = tmu_m_origin.pctag;
		if (((res_care_bits >> 3) & 1) == 0) 
			tmu_r.mwres = tmu_m_origin.mrtag;
		if (((res_care_bits >> 4) & 1) == 0) 
			tmu_r.op1res = tmu_m_origin.op1tag;
		if (((res_care_bits >> 5) & 1) == 0) 
			tmu_r.op2res = tmu_m_origin.op2tag;
		if (((res_care_bits >> 6) & 1) == 0) 
			tmu_r.op3res = tmu_m_origin.op3tag;
		return tuple2(tmu_allow, tmu_r);
        endmethod


	method Action tmu_unload(Bit#(8) slot);
		//clear all the M tables
		dyn_m_tptag.a.put(True, slot, dc_tag);
		dyn_m_citag.a.put(True, slot, dc_tag);
		dyn_m_ciop.a.put(True, slot, unpack(0));
		dyn_m_pctag.a.put(True, slot, dc_tag);
		dyn_m_auth.a.put(True, slot, dc_tag);
		dyn_m_op1tag.a.put(True, slot, dc_tag);
		dyn_m_op2tag.a.put(True, slot, dc_tag);
		dyn_m_op3tag.a.put(True, slot, dc_tag);
		dyn_m_mrtag.a.put(True, slot, dc_tag);

		// clear all the R tables
		dyn_r_op3res.a.put(True, slot, dc_tag);
		dyn_r_op2res.a.put(True, slot, dc_tag);
		dyn_r_op1res.a.put(True, slot, dc_tag);
		dyn_r_mwres.a.put(True, slot, dc_tag);
		dyn_r_pcres.a.put(True, slot, dc_tag);
		dyn_r_tpres.a.put(True, slot, dc_tag);
		dyn_r_auth.a.put(True, slot, dc_tag);

		// clear the hit count
		hit_counters.a.put(True, slot, unpack(0));
	endmethod


	method Action tmu_load(Bit#(9) slot, Bit#(5) field, TAG_A new_tag);

		if(field==0) begin
			dyn_m_tptag.b.put(True, truncate(slot), new_tag);
		end
		else if(field==1) begin
			dyn_m_citag.b.put(True, truncate(slot), new_tag);
		end
		else if(field==2) begin
			dyn_m_ciop.b.put(True, truncate(slot), unpack(truncate(pack(new_tag))));
		end
		else if(field==3) begin
			dyn_m_pctag.b.put(True, truncate(slot), new_tag);
		end
		else if(field==4) begin
			dyn_m_auth.b.put(True, truncate(slot), new_tag);
		end
		else if(field==5) begin
			dyn_m_op1tag.b.put(True, truncate(slot), new_tag);
		end
		else if(field==6) begin
			dyn_m_op2tag.b.put(True, truncate(slot), new_tag);
		end
		else if(field==7) begin
			dyn_m_op3tag.b.put(True, truncate(slot), new_tag);
		end
		else if(field==8) begin
			dyn_m_mrtag.b.put(True, truncate(slot), new_tag);
		end
		else if(field==9) begin
			dyn_r_op3res.b.put(True, truncate(slot), new_tag);
		end
		else if(field==10) begin
			dyn_r_op2res.b.put(True, truncate(slot), new_tag);
		end
		else if(field==11) begin
			dyn_r_op1res.b.put(True, truncate(slot), new_tag);
		end
		else if(field==12) begin
			dyn_r_mwres.b.put(True, truncate(slot), new_tag);
		end
		else if(field==13) begin
			dyn_r_pcres.b.put(True, truncate(slot), new_tag);
		end
		else if(field==14) begin
			dyn_r_tpres.b.put(True, truncate(slot), new_tag);
		end
		else if(field==15) begin
			dyn_r_auth.b.put(True, truncate(slot), new_tag);
		end
		else if(field==16) begin
			Bit#(8) tmp3 = unpack(truncate(pack(new_tag)));
			dyn_g1_table.b.put(True, slot, unpack(zeroExtend(tmp3)));
		end
		else if(field==17) begin
			dyn_g2_table.b.put(True, slot, unpack(0));
		end
	endmethod


	method Action tmu_rc_req(Bit#(8) slot);
		hit_counters.a.put(False, slot, ?);
	endmethod

	method Bit#(16) tmu_rc_read();
		Bit#(16) hit_count = hit_counters.a.read;
		return hit_count;
	endmethod

	/// increment the rule hit counter every cycle
	method Action tmu_rc_update(Bit#(16) new_count);
		hit_counters.b.put(True, dyn_mr_addr, new_count);
	endmethod

	//set the dont care tag value during bootstrapping..
	method Action set_dc_tag(TAG_A new_dc_tag);
		dc_tag <= new_dc_tag;
	endmethod

        method Action care_mem_load(Bit#(8) addr, Atom val);
		care_mem.b.put(True, addr, val);
        endmethod

        method Action res_care_load(Bit#(8) addr, Atom val);
		res_care_mem.b.put(True, addr, val);
        endmethod

        method Action gnd_g1_load(Bit#(9) addr, Atom val);
		gnd_g1_table.b.put(True, addr, val);
        endmethod

        method Action gnd_g2_load(Bit#(9) addr, Atom val);
		gnd_g2_table.b.put(True, addr, val);
        endmethod

        method Action gnd_mr_load(Bit#(8) addr, TMU_MR val);
	        gnd_m_tptag.b.put(True, addr, val.match_vector.tptag);
	        gnd_m_citag.b.put(True, addr, val.match_vector.citag);
	        gnd_m_ciop.b.put(True, addr, val.match_vector.opgrp);
	        gnd_m_pctag.b.put(True, addr, val.match_vector.pctag);
	        gnd_m_auth.b.put(True, addr, val.match_vector.auth);
	        gnd_m_op1tag.b.put(True, addr, val.match_vector.op1tag);
	        gnd_m_op2tag.b.put(True, addr, val.match_vector.op2tag);
	        gnd_m_op3tag.b.put(True, addr, val.match_vector.op3tag);
	        gnd_m_mrtag.b.put(True, addr, val.match_vector.mrtag);
	        gnd_r_op3res.b.put(True, addr, val.res_vector.op3res);
	        gnd_r_op2res.b.put(True, addr, val.res_vector.op2res);
                gnd_r_op1res.b.put(True, addr, val.res_vector.op1res);
	        gnd_r_mwres.b.put(True, addr, val.res_vector.mwres);
	        gnd_r_pcres.b.put(True, addr, val.res_vector.pcres);
	        gnd_r_tpres.b.put(True, addr, val.res_vector.tpres);
                gnd_r_auth.b.put(True, addr, val.res_vector.auth);
        endmethod

endmodule: mkTMU
