/************************************************************************************************
* Copyright  (c) 2011, 2012 University of Pennsylvania
*
* Permission to use, copy, modify, and distribute this software and
* its documentation for any purpose, without fee, and without a
* written agreement is hereby granted, provided that the above copyright
* notice and this paragraph and the following two paragraphs appear in
* all copies.
*
* IN NO EVENT SHALL THE UNIVERSITY OF PENNSYLVANIA BE LIABLE TO ANY PARTY FOR
* DIRECT, INDIRECT, SPECIAL, INCIDENTAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
* (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
* LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
* (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
* THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF PENNSYLVANIA
* HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*
* THE UNIVERSITY OF PENNSYLVANIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
* AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON
* AN "AS IS" BASIS, AND THE UNIVERSITY OF PENNSYLVANIA HAS NO OBLIGATIONS TO
* PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
***********************************************************************************************/

import SAFETypes::*;

interface TimerIfc;
        method ActionValue#(Tuple2#(Bool, TimerData)) do_timer (OPCODE co, TimerData src1);
        method Action set_slot (TimerData new_slot);
endinterface: TimerIfc

(*synthesize*)
module mkTimer# (parameter TimerData init_qt)(TimerIfc);

        Reg#(TimerReg) qt <- mkReg(TimerReg {meta: unpack(pack(MetaData{ag: INTEGER, tag: unpack(0)})), value: init_qt});
        Reg#(TimerReg) clk <- mkReg(TimerReg {meta: unpack(pack(MetaData{ag: INTEGER, tag: unpack(0)})), value: init_qt});
        Reg#(TimerReg) slot <- mkReg(TimerReg {meta: unpack(pack(MetaData{ag: INTEGER, tag: unpack(0)})), value: 0});

        function ActionValue#(Tuple2#(Bool, Bool)) check_time();
        actionvalue
	        TimerData new_clk = clk.value - 1;
	        Bool interrupt = False;
	        Bool clk_zero = False;
	        if (new_clk == 0) begin //if new clk value is 0
	                clk.value <= qt.value; //reload qt to clk
	                if (slot.value == 0) begin
	                        //perform timer interrupt if slot is 0. interrupt thread table will handle this if interrupt is true
	                        interrupt = True;
	                end
	                clk_zero = True;
	        end
	        else begin
	                clk.value <= new_clk; //decrement clk
	        end
	        return tuple2(interrupt, clk_zero);
        endactionvalue
        endfunction: check_time

        //handles all the timer calculation, and returns (isTimerInterrupt, slot_value)
        method ActionValue#(Tuple2#(Bool, TimerData)) do_timer (OPCODE co, TimerData src1);
                Tuple2#(Bool, Bool) res <- check_time;
                Bool interrupt = tpl_1(res);
                Bool clk_zero = tpl_2(res);
                TimerData slot_val = slot.value;
                case(co)
	                YIELD: return tuple2(interrupt, slot_val + 1);
	                GTIME: return tuple2(interrupt, slot_val + src1);
	                REMT: 	return tuple2(interrupt, slot_val - 1);
	                RDTIME: return tuple2(interrupt, slot_val);
	                default: begin
		                         if (clk_zero)
			                         return tuple2(interrupt, slot_val-1);
		                         else
			                         return tuple2(interrupt, slot_val);
		                 end
                endcase
        endmethod: do_timer

        //this method is used by commit unit to update slot
        method Action set_slot (TimerData new_slot);
                slot.value <= new_slot;
        endmethod: set_slot

endmodule: mkTimer
