/**********************************************
SAFE TestBench
Authors: Albert Kwon, Udit Dhawan
IC Group, University of Pennsylvania

Last Updated: 11.15.2011
Last change: changes made to create cycle and PC based breakpoints

		   : processor is now bootstrapped before the execution starts
		   : power on reset thread is located at 0x00 and contains the initial PC...
***********************************************/

package Top;

import SAFETypes::*;
import SAFE::*;
import Mem::*;
import IU::*;

import GetPut::*;
import ClientServer::*;
import Connectable::*;
import BRAM::*;

(*synthesize*)
module mkTop();

   	SAFEIfc proc <- mkSAFE;
   	MemIfc mem <- mkMem;
   	InterposedIfc iu <- mkIU;

   	Reg#(int) cycle <- mkReg(0);
   	Reg#(Bool) started <- mkReg(False);
   	Reg#(Bool) stream <- mkReg(False);
   	Reg#(A) addr <- mkReg('h154);	//10f
   	Reg#(Bool) running <- mkReg(False);

	int rf_depth = 32;
	int mem_depth = 1024;
	Reg#(Bit#(1)) phase <- mkReg(0);
	Reg#(int) stage <- mkReg(0);
	Reg#(int) mem_addr <- mkReg(0);
	Reg#(Bool) dumping <- mkReg(False);
	let fh <- mkReg(InvalidFile);

 	mkConnection(proc.store, mem.store);
   	mkConnection(proc.mem_rw, mem.proc_rw);

/* rules for the testbench */
   	rule start_processor (cycle == 1000);
		proc.start;
	  	running <= True;
                $display("SAFE powered up");
		$display("#Cycle  |  Authority  {PCtag PC_address}  {CItag CIop OP1 OP2 OP3 Imm}  {isLoadStore MemoryAddress}  {OP1.ag OP1.tag OP1.payload}  {OP2.ag OP2.tag OP2.payload}  {OP3.ag OP3.tag OP3.payload}  {MR.ag MR.tag MR.payload}  |  Error_Code");
   	endrule
        
        // rule setb (cycle == 2000);
        // 	proc.set_bp_cycle(1);
        //         $display("break");
        // endrule

        // rule cont (cycle == 3000);
        //         proc.clear_breakpoint;
        //         proc.enable_proc;
        //         $display("continue");
        // endrule
        
   	rule update_cycle;
      	        cycle <= cycle + 1;
   	endrule
        
        rule stream_load (cycle == 5000);
                mem.rw.request.put(BRAMRequest {write: True, responseOnWrite: False, address: ('h186) >> 2, datain: 
                                                CacheLine{a0: unpack('h300300000000008c0630c000000001c6), a1: unpack('h500f8000000000900000000000000001),
                                                          a2: unpack('h500f8000000000900000000000000001), a3: unpack('h500f8000000000900000000000000001)}});
                $display("writing stdin");
        endrule
        
   	rule is_running(running);
		let temp = proc.get_state(10);
                ThreadState proc_state = unpack(truncate(pack(temp)));
		if ((proc_state == HALTED) && (stage == 0)) begin
			started <= False;
			stage <= 1;
		end
		else if ((proc_state != HALTED) && (stage == 1)) begin
			started <= True;
			stage <= 2;
		end
		else if ((proc_state == HALTED) && (stage == 2)) begin
			running <= False;
			started <= False;
			dumping <= True;
		end
   	endrule

/*  	rule str_req((cycle != 0) && !stream && started);
      	mem.rw.request.put(iu.read_request(addr>>2));
      	stream <= True;
   	endrule

   	rule str_read((cycle != 0) && stream && started);
      	let tmp <- mem.rw.response.get;
      	if (!iu.do_stream(truncate(addr&3), False, tmp)) begin
	 		addr <= addr + 1;
			//mem.rw.request.put(iu.do_write(addr, unpack(6)));
			Atom err;
			Bit#(2) index = truncate(pack(addr&3));
			if(index==0) begin
				err = tmp.a0;
			end
			else if(index==1) begin
				err = tmp.a1;
			end
			else if(index==2) begin
				err = tmp.a2;
			end
			else begin
				err = tmp.a3;
			end
			//$display("PrivError: %x", err);
      	end
      	stream <= False;
   	endrule
*/
	// rule start_dumping(dumping && (mem_addr==0));
	// 	String dump_file = "memory.dump";
	// 	File prog <- $fopen(dump_file);

	// 	if(prog==InvalidFile) begin
	// 		//$display("Bad File!");
	// 		$finish(0);
	// 	end
	// 	else begin
	// 		//$display("\nCreating a dump file for memory contents...\n");
	// 		mem_addr <= 1;
	// 		fh <= prog;
	// 	end
	// endrule

	// rule dump_contents_phase0 (dumping && (mem_addr>0) && (mem_addr<=(rf_depth+mem_depth)) && (phase==0));
	// 	if(mem_addr==1) begin
	// 		$fwrite(fh, "Register File Dump: \n");
	// 	end
	// 	if(mem_addr <= rf_depth) begin
	// 		proc.get_reg_req(truncate(pack(mem_addr-1)));
	// 	end
	// 	if(mem_addr==(rf_depth+1)) begin
	// 		$fwrite(fh, "\nMemory Dump: \n \tMemory Address \t|\tAtom\n");
	// 	end
	// 	phase <= 1;
	// endrule

	// rule dump_contents_phase1 (dumping && (mem_addr>0) && (mem_addr<=(rf_depth+mem_depth)) && (phase==1));
	// 	if(mem_addr <= rf_depth) begin
	// 		Atom temp = proc.get_reg;
	// 		$fwrite(fh, "\tR%0d:\t%x\n", (mem_addr-1), temp);
	// 	end
	// 	mem_addr <= mem_addr + 1;
	// 	phase <= 0;
	// endrule

	// rule end_dumping (mem_addr > (rf_depth+mem_depth));
	// 	$fclose(fh);
	// 	//$display("\ncreated memory.dump");
	// 	$finish(0);
	// endrule

endmodule: mkTop

endpackage: Top
