/***************************************************
A small WriteBuffer built out of discrete registers for matching and a BRAM
Authors: Udit Dhawan
IC Group, University of Pennsylvania

Last Updated: 02.14.2012
***************************************************/

import SAFETypes::*;
import Vector::*;
import BRAM::*;
import BRAMCore::*;

import MatchItem::*;

interface WriteBufferIfc;
        method Action addr_match(A in_addr);
        method Action do_write(A in_addr, Atom in_val);
        method ActionValue#(Tuple2#(Bool, Atom)) do_read();
        method ActionValue#(A) flush_wb(Bit#(5) wb_addr);
        method ActionValue#(Tuple2#(Bool, Atom)) flush_wb_read(Bit#(5) wb_addr);
        method Bool is_empty();
        method Action reset_wb();
endinterface: WriteBufferIfc

(* synthesize *)
module mkWriteBuffer(WriteBufferIfc);

        //moved to lutrams
        Vector#(32, Reg#(Atom)) write_buffer <- replicateM(mkRegU);

        //vector of addresses of the stored atoms...
        Vector#(32, MatchItemIfc) match_items <- replicateM(mkMatchItem);
        Vector#(32, Reg#(Bool)) match_one_hot <- replicateM(mkReg(False));

        Reg#(Bit#(5)) addr <- mkReg(0);
        Reg#(Bit#(5)) head <- mkReg(0);
        Reg#(Bit#(5)) tail <- mkReg(0);
        Reg#(Bit#(5)) num_items <- mkReg(0);
        Reg#(Bool) found <- mkReg(False);


        method Action addr_match(A in_addr);
                match_one_hot[0] <= match_items[0].do_match(in_addr);
                match_one_hot[1] <= match_items[1].do_match(in_addr);
                match_one_hot[2] <= match_items[2].do_match(in_addr);
                match_one_hot[3] <= match_items[3].do_match(in_addr);
                match_one_hot[4] <= match_items[4].do_match(in_addr);
                match_one_hot[5] <= match_items[5].do_match(in_addr);
                match_one_hot[6] <= match_items[6].do_match(in_addr);
                match_one_hot[7] <= match_items[7].do_match(in_addr);
                match_one_hot[8] <= match_items[8].do_match(in_addr);
                match_one_hot[9] <= match_items[9].do_match(in_addr);
                match_one_hot[10] <= match_items[10].do_match(in_addr);
                match_one_hot[11] <= match_items[11].do_match(in_addr);
                match_one_hot[12] <= match_items[12].do_match(in_addr);
                match_one_hot[13] <= match_items[13].do_match(in_addr);
                match_one_hot[14] <= match_items[14].do_match(in_addr);
                match_one_hot[15] <= match_items[15].do_match(in_addr);
                match_one_hot[16] <= match_items[16].do_match(in_addr);
                match_one_hot[17] <= match_items[17].do_match(in_addr);
                match_one_hot[18] <= match_items[18].do_match(in_addr);
                match_one_hot[19] <= match_items[19].do_match(in_addr);
                match_one_hot[20] <= match_items[20].do_match(in_addr);
                match_one_hot[21] <= match_items[21].do_match(in_addr);
                match_one_hot[22] <= match_items[22].do_match(in_addr);
                match_one_hot[23] <= match_items[23].do_match(in_addr);
                match_one_hot[24] <= match_items[24].do_match(in_addr);
                match_one_hot[25] <= match_items[25].do_match(in_addr);
                match_one_hot[26] <= match_items[26].do_match(in_addr);
                match_one_hot[27] <= match_items[27].do_match(in_addr);
                match_one_hot[28] <= match_items[28].do_match(in_addr);
                match_one_hot[29] <= match_items[29].do_match(in_addr);
                match_one_hot[30] <= match_items[30].do_match(in_addr);
                match_one_hot[31] <= match_items[31].do_match(in_addr);
        endmethod

        //return (match_found, atom)
        method ActionValue#(Tuple2#(Bool, Atom)) do_read();
                Bit#(5) match_addr = 0;
                Bool match_found = False;

                //convert one_hot encoded match address to log (cam_depth)-bit encoded
                if (match_one_hot[0])  begin
                        match_addr = 0;
                        match_found = True;
                end
                else if (match_one_hot[1]) begin
                        match_addr = 1;
                        match_found = True;
                end
                else if (match_one_hot[2]) begin
                        match_addr = 2;
                        match_found = True;
                end
                else if (match_one_hot[3]) begin
                        match_addr = 3;
                        match_found = True;
                end
                else if (match_one_hot[4]) begin
                        match_addr = 4;
                        match_found = True;
                end
                else if (match_one_hot[5]) begin
                        match_addr = 5;
                        match_found = True;
                end
                else if (match_one_hot[6]) begin
                        match_addr = 6;
                        match_found = True;
                end
                else if (match_one_hot[7]) begin
                        match_addr = 7;
                        match_found = True;
                end
                else if (match_one_hot[8]) begin
                        match_addr = 8;
                        match_found = True;
                end
                else if (match_one_hot[9]) begin
                        match_addr = 9;
                        match_found = True;
                end
                else if (match_one_hot[10]) begin
                        match_addr = 10;
                        match_found = True;
                end
                else if (match_one_hot[11]) begin
                        match_addr = 11;
                        match_found = True;
                end
                else if (match_one_hot[12]) begin
                        match_addr = 12;
                        match_found = True;
                end
                else if (match_one_hot[13]) begin
                        match_addr = 13;
                        match_found = True;
                end
                else if (match_one_hot[14]) begin
                        match_addr = 14;
                        match_found = True;
                end
                else if (match_one_hot[15]) begin
                        match_addr = 15;
                        match_found = True;
                end
                else if (match_one_hot[16])  begin
                        match_addr = 16;
                        match_found = True;
                end
                else if (match_one_hot[17]) begin
                        match_addr = 17;
                        match_found = True;
                end
                else if (match_one_hot[18]) begin
                        match_addr = 18;
                        match_found = True;
                end
                else if (match_one_hot[19]) begin
                        match_addr = 19;
                        match_found = True;
                end
                else if (match_one_hot[20]) begin
                        match_addr = 20;
                        match_found = True;
                end
                else if (match_one_hot[21]) begin
                        match_addr = 21;
                        match_found = True;
                end
                else if (match_one_hot[22]) begin
                        match_addr = 22;
                        match_found = True;
                end
                else if (match_one_hot[23]) begin
                        match_addr = 23;
                        match_found = True;
                end
                else if (match_one_hot[24]) begin
                        match_addr = 24;
                        match_found = True;
                end
                else if (match_one_hot[25]) begin
                        match_addr = 25;
                        match_found = True;
                end
                else if (match_one_hot[26]) begin
                        match_addr = 26;
                        match_found = True;
                end
                else if (match_one_hot[27]) begin
                        match_addr = 27;
                        match_found = True;
                end
                else if (match_one_hot[28]) begin
                        match_addr = 28;
                        match_found = True;
                end
                else if (match_one_hot[29]) begin
                        match_addr = 29;
                        match_found = True;
                end
                else if (match_one_hot[30]) begin
                        match_addr = 30;
                        match_found = True;
                end
                else if (match_one_hot[31]) begin
                        match_addr = 31;
                        match_found = True;
                end

                Atom val = write_buffer[match_addr];
                addr <= match_addr;
                found <= match_found;
                return tuple2(match_found, val);
        endmethod

        method Action do_write(A in_addr, Atom in_val);
                if (found) begin
                        write_buffer[addr] <= in_val;
                        found <= False;
                end
                else begin
                        match_items[tail].write(in_addr);
                        write_buffer[tail] <= in_val;
                        tail <= tail + 1;
                end
        endmethod

        method ActionValue#(A) flush_wb(Bit#(5) wb_addr);
                A mem_addr = match_items[wb_addr].fetch_item;
                return mem_addr;
        endmethod

        method ActionValue#(Tuple2#(Bool, Atom)) flush_wb_read(Bit#(5) wb_addr);
                Atom val = write_buffer[wb_addr];
                Bool is_valid = match_items[wb_addr].is_valid;
                return tuple2(is_valid, val);
        endmethod

        method Action reset_wb();
                head <= 0;
                tail <= 0;
                addr <= 0;
                num_items <= 0;

                match_items[0].clear_item;
                match_items[1].clear_item;
                match_items[2].clear_item;
                match_items[3].clear_item;
                match_items[4].clear_item;
                match_items[5].clear_item;
                match_items[6].clear_item;
                match_items[7].clear_item;
                match_items[8].clear_item;
                match_items[9].clear_item;
                match_items[10].clear_item;
                match_items[11].clear_item;
                match_items[12].clear_item;
                match_items[13].clear_item;
                match_items[14].clear_item;
                match_items[15].clear_item;
                match_items[16].clear_item;
                match_items[17].clear_item;
                match_items[18].clear_item;
                match_items[19].clear_item;
                match_items[20].clear_item;
                match_items[21].clear_item;
                match_items[22].clear_item;
                match_items[23].clear_item;
                match_items[24].clear_item;
                match_items[25].clear_item;
                match_items[26].clear_item;
                match_items[27].clear_item;
                match_items[28].clear_item;
                match_items[29].clear_item;
                match_items[30].clear_item;
                match_items[31].clear_item;
        endmethod

        method Bool is_empty();
                return tail == head;
        endmethod

endmodule
